"""RAS Class and functions."""

import glob
import logging
import os
import platform
import re
import subprocess
import time
import warnings
from pathlib import Path
from typing import List

import fiona
import geopandas as gpd
import h5py
import pandas as pd

try:
    import pythoncom
except SystemError:
    warnings.warn("Windows OS is required to run ripple1d. Many features will not work on other OS's.")

from pyproj import CRS

from ripple1d.consts import (
    FLOW_HDF_PATH,
    NORMAL_DEPTH,
    PROFILE_NAMES_HDF_PATH,
    SHOW_RAS,
    SUPPORTED_LAYERS,
    TERRAIN_NAME,
    TERRAIN_PATH,
    WSE_HDF_PATH,
    XS_NAMES_HDF_PATH,
)
from ripple1d.data_model import FlowChangeLocation, Junction, Reach
from ripple1d.errors import (
    FlowTitleAlreadyExistsError,
    HECRASVersionNotInstalledError,
    NoCrossSectionLayerError,
    NoFlowFileSpecifiedError,
    NoGeometryFileSpecifiedError,
    NoRiverLayerError,
    PlanTitleAlreadyExistsError,
    RASComputeTimeoutError,
)

# ToManyPlansError,
from ripple1d.rasmap import PLAN, RASMAP_631, TERRAIN
from ripple1d.utils.dg_utils import get_terrain_exe_path
from ripple1d.utils.ripple_utils import (
    assert_no_mesh_error,
    assert_no_ras_compute_error_message,
    assert_no_ras_geometry_error,
    assert_no_store_all_maps_error_message,
    decode,
    replace_line_in_contents,
    search_contents,
    text_block_from_start_end_str,
)

if platform.system() == "Windows":
    import win32com.client
    from pythoncom import com_error


RAS_FILE_TYPES = ["Plan", "Flow", "Geometry", "Project"]

VALID_PLANS = [f".p{i:02d}" for i in range(1, 100)] + [f".P{i:02d}" for i in range(1, 100)]
VALID_GEOMS = [f".g{i:02d}" for i in range(1, 100)] + [f".G{i:02d}" for i in range(1, 100)]
VALID_STEADY_FLOWS = [f".f{i:02d}" for i in range(1, 100)] + [f".F{i:02d}" for i in range(1, 100)]
VALID_UNSTEADY_FLOWS = [f".u{i:02d}" for i in range(1, 100)] + [f".U{i:02d}" for i in range(1, 100)]
VALID_QUASISTEADY_FLOWS = [f".q{i:02d}" for i in range(1, 100)] + [f".Q{i:02d}" for i in range(1, 100)]


# Decorator Functions
def check_crs(func):
    """Check CRS decorator."""

    def wrapper(self, *args, **kwargs):
        if self.crs is None:
            raise ValueError("Projection cannot be None")
        return func(self, *args, **kwargs)

    return wrapper


def combine_root_extension(func):
    """Combine root extension decorator."""

    def wrapper(self, *args, **kwargs):
        extensions = func(self, *args, **kwargs)
        if isinstance(extensions, list):
            return [
                self._ras_root_path + "." + extension.replace(" ", "").lstrip(".").lower() for extension in extensions
            ]
        else:
            return self._ras_root_path + "." + extensions.replace(" ", "").lstrip(".").lower()

    return wrapper


def check_version_installed(version: str):
    """Check version installed decorator."""

    def decorator(func):
        def wrapper(self, *args, **kwargs):
            try:
                assert win32com.client.Dispatch(f"RAS{version}.HECRASCONTROLLER", pythoncom.CoInitialize())
                self.version = version
            except com_error:
                raise HECRASVersionNotInstalledError(
                    f"Could not find the specified RAS version; please ensure it is installed. Version provided: {version}."
                )
            return func(self, *args, **kwargs)

        return wrapper

    return decorator


def check_windows(func):
    """Check windows decorator."""

    def wrapper(self, *args, **kwargs):
        if platform.system() != "Windows":
            raise SystemError("This method can only be run on a Windows machine.")
        return func(self, *args, **kwargs)

    return wrapper


def add_fid_index(func):
    """Add FID index decorator."""

    def wrapper(*args, **kwargs):
        gdf = func(*args, **kwargs)
        gdf.index.name = "ID"
        return gdf

    return wrapper


# classes
class RasManager:
    """Manage HEC-RAS projects."""

    def __init__(
        self,
        ras_text_file_path: str,
        version: str = "631",
        terrain_path: str = None,
        crs: CRS = None,
        new_project: bool = False,
    ):
        self.version = version
        self.terrain_path = terrain_path
        self.ras_project = RasProject(ras_text_file_path, new_file=new_project)

        self.crs = CRS(crs)
        self.plans = self.get_plans()
        self.geoms = self.get_geoms()
        self.flows = self.get_flows()
        self.plan = self.current_plan

    def __repr__(self):
        """Representation of the RasManager class."""
        return f"RasManager(project={self.ras_project._ras_text_file_path} ras-version={self.version})"

    @classmethod
    def from_gpkg(
        cls,
        ras_project_text_file: str,
        nwm_id,
        ras_gpkg_file_path: str,
        version: str = "631",
        terrain_path: str = None,
    ):
        """Create a new RasManager object from a geopackage."""
        inst = cls(
            ras_project_text_file,
            version,
            terrain_path=terrain_path,
            crs=gpd.read_file(ras_gpkg_file_path, layer="XS").crs,
            new_project=True,
        )

        inst.new_geom_from_gpkg(ras_gpkg_file_path, nwm_id)
        inst.ras_project.write_contents()
        return inst

    @property
    def current_plan(self):
        """Get the current plan."""
        for plan in self.plans.values():
            if plan.file_extension == self.ras_project.current_plan:
                return plan

    @property
    def projection_file(self):
        """Write the current projection to file and return the file path."""
        projection_file = os.path.join(self.ras_project._ras_dir, "crs.prj")
        with open(projection_file, "w") as f:
            f.write(self.crs.to_wkt("WKT1_ESRI"))
        return projection_file

    def get_plans(self):
        """Create plan objects for each plan."""
        plans = {}
        for plan_file in self.ras_project.plans:
            try:
                plan = RasPlanText(plan_file, self.crs)
                plans[plan.title] = plan
            except FileNotFoundError:
                logging.info(f"Could not find plan file: {plan_file}")
        return plans

    @check_crs
    def get_geoms(self):
        """Create geom objects for each geom."""
        geoms = {}
        for geom_file in self.ras_project.geoms:
            try:
                geom = RasGeomText(geom_file, self.crs)
                geoms[geom.title] = geom
            except FileNotFoundError:
                logging.warning(f"Could not find geom file: {geom_file}")
        return geoms

    def get_flows(self):
        """Create flow objects for each flow."""
        flows = {}
        for flow_file in self.ras_project.steady_flows:
            try:
                flow = RasFlowText(flow_file)
                flows[flow.title] = flow
            except FileNotFoundError:
                logging.warning(f"Could not find flow file: {flow_file}")
        return flows

    @check_windows
    @check_version_installed("631")
    def run_sim(
        self,
        pid_running=None,
        close_ras=True,
        show_ras=False,
        ignore_store_all_maps_error: bool = False,
        timeout_seconds=None,
    ):
        """
        Run the current plan.

        Args:
            pid_running (_type_, optional): _description_. Defaults to None.
            close_ras (bool, optional): boolean to close RAS or not after computing. Defaults to True.
            show_ras (bool, optional): boolean to show RAS or not when computing. Defaults to False.
        """
        compute_message_file = self.ras_project._ras_root_path + f"{self.plan.file_extension}.computeMsgs.txt"

        RC = win32com.client.Dispatch(f"RAS{self.version}.HECRASCONTROLLER", pythoncom.CoInitialize())
        try:
            RC.Project_Open(self.ras_project._ras_text_file_path)
            if show_ras:
                RC.ShowRas()

            RC.Compute_CurrentPlan()
            deadline = (timeout_seconds + time.time()) if timeout_seconds else float("inf")
            while not RC.Compute_Complete():
                if time.time() > deadline:
                    raise RASComputeTimeoutError(
                        f"timed out computing current plan for RAS project: {self.ras_project._ras_text_file_path}"
                    )
                # must keep checking for mesh errors while RAS is running
                # because mesh error will cause blocking popup message
                assert_no_mesh_error(compute_message_file, require_exists=False)  # file might not exist immediately
                time.sleep(0.2)
            time.sleep(
                3
            )  # ample sleep to account for race condition of RAS flushing final lines to compute_message_file

            assert_no_mesh_error(compute_message_file, require_exists=True)
            assert_no_ras_geometry_error(compute_message_file)
            assert_no_ras_compute_error_message(compute_message_file)
            if not ignore_store_all_maps_error:
                assert_no_store_all_maps_error_message(compute_message_file)
        finally:
            if close_ras:
                RC.Project_Close()
                RC.QuitRas()

    def normal_depth_run(
        self,
        plan_flow_title: str,
        geom_title: str,
        flow_change_locations: list[FlowChangeLocation],
        profile_names: list[str],
        normal_depth: float = NORMAL_DEPTH,
        write_depth_grids: bool = False,
        show_ras: bool = False,
        run_ras: bool = True,
    ):
        """Create a new normal depth run."""
        if plan_flow_title in self.flows.keys():
            raise FlowTitleAlreadyExistsError(f"The specified flow title {plan_flow_title} already exists")

        # get a new extension number for the new flow file
        new_extension_number = get_new_extension_number(self.flows)
        flow_text_file = self.ras_project._ras_root_path + f".f{new_extension_number}"

        # create new flow
        rft = RasFlowText(flow_text_file, new_file=True)

        # write headers
        rft.contents += rft.write_headers(plan_flow_title, profile_names)

        for fcl in flow_change_locations:
            # write discharges
            rft.contents += rft.write_discharges(fcl.flows, fcl.river, fcl.reach, fcl.rs)

        for fcl in flow_change_locations:
            # write normal depth
            rft.contents += rft.write_ds_normal_depth(len(fcl.flows), normal_depth, fcl.river, fcl.reach)

        # write flow file content
        rft.write_contents()

        # add new flow to the ras class
        self.flows[plan_flow_title] = rft
        self.flow = rft

        # add to ras project contents
        self.ras_project.contents.append(f"Flow File=f{new_extension_number}")

        self.write_new_plan_text_file(plan_flow_title, geom_title, write_depth_grids, show_ras, run_ras)

    def kwses_run(
        self,
        plan_flow_title: str,
        geom_title: str,
        depths: List[float],
        wses: List[float],
        flows: List[float],
        river: str,
        reach: str,
        us_river_station: float,
        write_depth_grids: bool = False,
        show_ras: bool = False,
        run_ras: bool = True,
    ):
        """Create a new known water surface elevation run."""
        if plan_flow_title in self.flows.keys():
            raise FlowTitleAlreadyExistsError(f"The specified flow title {plan_flow_title} already exists")

        # get a new extension number for the new flow file
        new_extension_number = get_new_extension_number(self.flows)
        flow_text_file = self.ras_project._ras_root_path + f".f{new_extension_number}"

        # create new flow
        rft = RasFlowText(flow_text_file, new_file=True)

        profile_names = [f"f_{flow}-z_{str(wse).replace('.','_')}" for flow, wse in zip(flows, wses)]

        # write headers
        rft.contents += rft.write_headers(plan_flow_title, profile_names)

        # write discharges
        rft.contents += rft.write_discharges(flows, river, reach, us_river_station)

        # write DS boundary conditions
        rft.contents += rft.write_ds_known_wse(wses, river, reach)

        # write flow file content
        rft.write_contents()

        # add new flow to the ras class
        self.flows[plan_flow_title] = rft
        self.flow = rft

        # add to ras project contents
        self.ras_project.contents.append(f"Flow File=f{new_extension_number}")

        self.write_new_plan_text_file(plan_flow_title, geom_title, write_depth_grids, show_ras, run_ras)

    def new_geom_from_gpkg(
        self,
        ras_gpkg_file_path: str,
        title: str,
    ):
        """Create a new geometry file from a geopackage."""
        new_extension_number = get_new_extension_number(self.geoms)
        text_file = self.ras_project._ras_root_path + f".g{new_extension_number}"
        geom_text_file = RasGeomText.from_gpkg(ras_gpkg_file_path, title, self.version, text_file)
        geom_text_file.write_contents()
        self.geoms[geom_text_file.title] = geom_text_file
        self.ras_project.contents.append(f"Geom File=g{new_extension_number}")

    def update_rasmapper_for_mapping(self):
        """Write a rasmapper file to output depth grids for the current plan."""
        # manage rasmapper
        map_file = f"{self.ras_project._ras_root_path}.rasmap"

        if os.path.exists(map_file):
            os.remove(map_file)

        if os.path.exists(map_file + ".backup"):
            os.remove(map_file + ".backup")

        terrain_relative_path = os.path.relpath(self.terrain_path, self.ras_project._ras_dir)
        terrain_name = os.path.splitext(os.path.basename(self.terrain_path))[0]

        rasmap = RasMap(map_file, self.plan.geom, self.version)
        rasmap.update_crs(self.projection_file)
        rasmap.add_terrain(terrain_name, terrain_relative_path)
        rasmap.add_plan_layer(
            self.plan.title,
            os.path.basename(self.plan.hdf_file),
            self.plan.flow.profile_names,
        )
        rasmap.add_result_layers(self.plan.title, self.plan.flow.profile_names, "Depth")
        rasmap.write()

    def write_new_plan_text_file(
        self, plan_flow_title, geom_title, write_depth_grids: bool = False, show_ras=False, run_ras=True
    ):
        """Write new plan text file decorator."""
        if plan_flow_title in self.plans.keys():
            raise PlanTitleAlreadyExistsError(f"The specified plan title {plan_flow_title} already exists")

        if plan_flow_title not in self.flows.keys():
            raise ValueError(f"The specified flow title {plan_flow_title} does not exist")

        if geom_title not in self.geoms.keys():
            raise ValueError(f"The specified geom title {geom_title} does not exist")

        # get a new extension number for the new plan
        new_extension_number = get_new_extension_number(self.plans)

        plan_text_file = self.ras_project._ras_root_path + f".p{new_extension_number}"

        # create plan
        rpt = RasPlanText(plan_text_file, self.crs, new_file=True)

        # populate new plan info
        rpt.new_plan_contents(
            plan_flow_title,
            plan_flow_title,
            self.flows[plan_flow_title],
            self.geoms[geom_title],
            write_depth_grids,
        )

        # write content
        rpt.write_contents()

        # add new plan to the ras class
        self.plans[plan_flow_title] = rpt
        self.plan = rpt

        # add to ras project contents
        self.ras_project.contents.append(f"Plan File=p{new_extension_number}")

        # update the content of the RAS project file
        self.contents = self.ras_project.set_current_plan(self.plans[plan_flow_title].file_extension)

        # write the update RAS project file content
        self.ras_project.write_updated_contents()

        if write_depth_grids:
            self.update_rasmapper_for_mapping()

        if run_ras:
            # run the RAS plan
            self.run_sim(close_ras=True, show_ras=show_ras, ignore_store_all_maps_error=True)


class RasTextFile:
    """Represents a HEC-RAS text file."""

    def __init__(self, ras_text_file_path, new_file=False):
        self._ras_text_file_path = ras_text_file_path

        if not new_file and not os.path.exists(ras_text_file_path):
            raise FileNotFoundError(f"could not find {ras_text_file_path}")
        else:
            self._ras_text_file_path = ras_text_file_path
            self._ras_root_path = os.path.splitext(self._ras_text_file_path)[0]

        if not new_file:
            self.read_contents()
        else:
            self.contents = []

    def __repr__(self):
        """Representation of the RasTextFile class."""
        return f"RasTextFile({self._ras_text_file_path})"

    def read_contents(self):
        """Read the contents of the text file."""
        if not os.path.exists(self._ras_text_file_path):
            raise FileNotFoundError(f"could not find {self._ras_text_file_path}")
        with open(self._ras_text_file_path) as f:
            self.contents = f.read().splitlines()

    def write_contents(self):
        """Write the contents of the text file."""
        if os.path.exists(self._ras_text_file_path):
            raise FileExistsError(f"The specified file already exists {self._ras_text_file_path}")

        logging.info(f"writing: {self._ras_text_file_path}")
        with open(self._ras_text_file_path, "w") as f:
            f.write("\n".join(self.contents))

    def write_updated_contents(self):
        """Write the updated contents of the text file."""
        if not os.path.exists(self._ras_text_file_path):
            raise FileNotFoundError(f"The specified file doesn't exists {self._ras_text_file_path}")

        logging.info(f"updating: {self._ras_text_file_path}")
        with open(self._ras_text_file_path, "w") as f:
            f.write("\n".join(self.contents))

    @property
    def file_extension(self):
        """Get the file extension."""
        return Path(self._ras_text_file_path).suffix


class RasProject(RasTextFile):
    """Represents a HEC-RAS project file."""

    def __init__(self, ras_text_file_path: str, new_file: bool = False):
        super().__init__(ras_text_file_path, new_file)

        if self.file_extension != ".prj":
            raise TypeError(f"Project extenstion must be .prj, not {self.file_extension}")

        self._ras_project_basename = os.path.splitext(os.path.basename(self._ras_text_file_path))[0]
        self._ras_dir = os.path.dirname(self._ras_text_file_path)
        os.makedirs(self._ras_dir, exist_ok=True)

        if new_file:
            self.contents = [
                f"Proj Title={self._ras_project_basename}",
                "Current Plan=",
            ]

    def __repr__(self):
        """Representation of the RasProject class."""
        return f"RasProject({self._ras_text_file_path})"

    @classmethod
    def from_str(cls, text_string: str, ras_text_file_path: str = ""):
        """Initiate a RasProject class from a string."""
        inst = cls(ras_text_file_path, new_file=True)
        inst.contents = text_string.splitlines()
        return inst

    @property
    def title(self):
        """Title of the HEC-RAS project."""
        return search_contents(self.contents, "Proj Title")

    @property
    def units(self):
        """Units of the HEC-RAS project."""
        if "English Units" in self.contents:
            return "English"
        else:
            return "Metric"

    @property
    @combine_root_extension
    def plans(self):
        """Get the plans associated with this project."""
        return [f".{ext}" for ext in search_contents(self.contents, "Plan File", expect_one=False)]

    @property
    @combine_root_extension
    def geoms(self):
        """Get the geometry files associated with this project."""
        return search_contents(self.contents, "Geom File", expect_one=False)

    @property
    @combine_root_extension
    def unsteady_flows(self):
        """Get the unsteady flow files associated with this project."""
        return search_contents(self.contents, "Unsteady File", expect_one=False)

    @property
    @combine_root_extension
    def steady_flows(self):
        """Get the steady flow files associated with this project."""
        return search_contents(self.contents, "Flow File", expect_one=False)

    @property
    def n_geoms(self):
        """Get the number of geometry files associated with this project."""
        return len(self.geoms)

    @property
    def n_plans(self):
        """Get the number of plans associated with this project."""
        return len(self.plans)

    @property
    def n_flows(self):
        """Get the number of flow files associated with this project."""
        return len(self.steady_flows)

    @property
    def current_plan(self):
        """Get the current plan."""
        return f".{search_contents(self.contents, 'Current Plan')}"

    def set_current_plan(self, plan_ext):
        """
        Return new contents with the specified plan as the current RAS plan.

        Args:
            plan_ext: The plan extension to set as the current plan
        """
        new_contents = self.contents
        if f"{plan_ext}" not in VALID_PLANS:
            raise TypeError(f"Plan extenstion must be one of .p01-.p99, not {plan_ext}")
        else:
            new_contents = replace_line_in_contents(new_contents, "Current Plan", plan_ext.lstrip("."))

        # TODO: Update this to put it with the other plans
        if f"Plan File={plan_ext.lstrip('.')}" not in new_contents:
            new_contents.append(f"Plan File={plan_ext.lstrip('.')}")
        logging.info("set plan!")
        return new_contents


class RasPlanText(RasTextFile):
    """Represents a HEC-RAS plan file."""

    def __init__(self, ras_text_file_path: str, crs: str = None, new_file: bool = False):
        super().__init__(ras_text_file_path, new_file)
        if self.file_extension not in VALID_PLANS:
            raise TypeError(f"Plan extenstion must be one of .p01-.p99, not {self.file_extension}")
        self.crs = crs
        self.hdf_file = self._ras_text_file_path + ".hdf"

    def __repr__(self):
        """Representation of the RasPlanText class."""
        return f"RasPlanText({self._ras_text_file_path})"

    @classmethod
    def from_str(cls, text_string: str, crs, ras_text_file_path: str = ""):
        """Initiate a RasPlanText class from a string."""
        inst = cls(ras_text_file_path, crs, new_file=True)
        inst.contents = text_string.splitlines()
        return inst

    @property
    def title(self):
        """Title of this HEC-RAS plan."""
        return search_contents(self.contents, "Plan Title")

    @property
    def version(self):
        """HEC-RAS version."""
        return search_contents(self.contents, "Program Version")

    @property
    @combine_root_extension
    def plan_geom_file(self):
        """Geometry flow file associated with this plan."""
        return self.plan_geom_extension

    @property
    @combine_root_extension
    def plan_unsteady_flow_file(self):
        """Unsteady flow file associated with this plan."""
        return self.plan_unsteady_extension

    @property
    @combine_root_extension
    def plan_steady_file(self):
        """Steady flow file associated with this plan."""
        return self.plan_steady_extension

    @property
    def plan_geom_extension(self):
        """Geometry extension associated with this plan."""
        try:
            return f".{search_contents(self.contents, 'Geom File')}"
        except ValueError:
            raise NoGeometryFileSpecifiedError(
                f"Could not find a specified geometry file for plan: {self.title} | {self._ras_text_file_path}"
            )

    @property
    def plan_unsteady_extension(self):
        """Unsteady flow extension associated with this plan."""
        return f".{search_contents(self.contents, 'Unsteady File')}"

    @property
    def plan_steady_extension(self):
        """Steady flow extension associated with this plan."""
        try:
            return f".{search_contents(self.contents, 'Flow File')}"
        except ValueError:
            raise NoFlowFileSpecifiedError(
                f"Could not find a specified flow file for plan: {self.title} | {self._ras_text_file_path}"
            )

    @property
    @check_crs
    def geom(self):
        """Represents the HEC-RAS geometry file associated with this plan."""
        return RasGeomText(self.plan_geom_file, self.crs)

    @property
    def flow(self):
        """Represents the HEC-RAS flow file associated with this plan."""
        return RasFlowText(self.plan_steady_file)

    def new_plan_from_existing(self, title, short_id, geom_ext, flow_ext):
        """Populate the content of the new plan with basic attributes (title, short_id, flow, and geom)."""
        new_contents = self.contents
        if len(title) > 80:
            raise ValueError("Short Identifier must be less than 80 characters")
        else:
            new_contents = replace_line_in_contents(new_contents, "Plan Title", title)

        if len(short_id) > 80:
            raise ValueError("Short Identifier must be less than 80 characters")
        else:
            new_contents = replace_line_in_contents(new_contents, "Short Identifier", short_id)

        if f".{geom_ext}" not in VALID_GEOMS:
            raise TypeError(f"Geometry extenstion must be one of g01-g99, not {geom_ext}")
        else:
            new_contents = replace_line_in_contents(new_contents, "Geom File", geom_ext)

        if f".{flow_ext}" not in VALID_STEADY_FLOWS:
            raise TypeError(f"Flow extenstion must be one of f01-f99, not {flow_ext}")
        else:
            new_contents = replace_line_in_contents(new_contents, "Flow File", flow_ext)

        try:
            new_contents = replace_line_in_contents(new_contents, "Run RASMapper", "-1")
        except ValueError:
            new_contents.append("Run RASMapper=-1 ")

        return new_contents

    def new_plan_contents(self, title: str, short_id: str, flow, geom, run_rasmapper: bool = False):
        """
        Populate the content of the plan with basic attributes (title, short_id, flow, and geom).

        Raises
        ------
            RuntimeError: raise run time error if the plan already has content associated with it
        """
        if self.contents:
            raise RuntimeError(f"content already exists for this plan: {self._ras_text_file_path}")

        # create necessary lines for the content of the plan text file.
        if len(title) > 80:
            raise ValueError("Short Identifier must be less than 80 characters")
        else:
            self.contents.append(f"Plan Title={title}")
        if len(short_id) > 80:
            raise ValueError("Short Identifier must be less than 80 characters")
        else:
            self.contents.append(f"Short Identifier={short_id}")

        if f"{geom.file_extension}" not in VALID_GEOMS:
            raise TypeError(f"Geometry extenstion must be one of .g01-.g99, not {geom.file_extension}")
        else:
            self.contents.append(f"Geom File={geom.file_extension.lstrip('.')}")

        if f"{flow.file_extension}" not in VALID_STEADY_FLOWS:
            raise TypeError(f"Flow extenstion must be one of .f01-.f99, not {flow.file_extension}")
        else:
            self.contents.append(f"Flow File={flow.file_extension.lstrip('.')}")
        if run_rasmapper:
            self.contents.append("Run RASMapper=-1 ")
        else:
            self.contents.append("Run RASMapper= 0 ")

    def read_rating_curves(self) -> dict:
        """
        Read the flow and water surface elevations resulting from the computed plan.

        Raises
        ------
            FileNotFoundError: _description_

        Returns
        -------
            dict: A dictionary containing "wse" and "flow" keys whose values are pandas dataframes
        """
        # check if the hdf file exists; raise error if it does not
        if not self.hdf_file:
            self.hdf_file = self.text_file + ".hdf"

            if not os.path.exists(self.hdf_file):
                raise FileNotFoundError(f'The file "{self.hdf_file}" does not exists')

        def remove_multiple_spaces(x):
            return re.sub(" +", " ", x)

        # read the hdf file
        with h5py.File(self.hdf_file) as hdf:
            # get columns and indexes for the wse and flow arrays
            columns = decode(pd.DataFrame(hdf[XS_NAMES_HDF_PATH]))
            index = decode(pd.DataFrame(hdf[PROFILE_NAMES_HDF_PATH]))

            # remove multiple spaces in between the river-reach-riverstation ids.
            columns[0] = columns[0].apply(remove_multiple_spaces)

            # create dataframes for the wse and flow results
            wse = pd.DataFrame(hdf.get(WSE_HDF_PATH), columns=columns[0].values, index=index[0].values).T
            flow = pd.DataFrame(hdf.get(FLOW_HDF_PATH), columns=columns[0].values, index=index[0].values).T

        return wse, flow


class RasGeomText(RasTextFile):
    """Represents a HEC-RAS geometry text file."""

    def __init__(self, ras_text_file_path: str, crs: str = None, new_file=False):
        super().__init__(ras_text_file_path, new_file)
        if not new_file and self.file_extension not in VALID_GEOMS:
            raise TypeError(f"Geometry extenstion must be one of .g01-.g99, not {self.file_extension}")

        self.crs = CRS(crs)
        self.hdf_file = self._ras_text_file_path + ".hdf"

    def __repr__(self):
        """Representation of the RasGeomText class."""
        return f"RasGeomText({self._ras_text_file_path})"

    @classmethod
    def from_str(cls, text_string: str, crs, ras_text_file_path: str = ""):
        """Initiate the RASGeomText class from a string."""
        inst = cls(ras_text_file_path, crs, new_file=True)
        inst.contents = text_string.splitlines()
        return inst

    @classmethod
    def from_gpkg(cls, gpkg_path, title: str, version: str, ras_text_file_path: str = ""):
        """Initiate the RASGeomText class from a geopackage."""
        inst = cls(ras_text_file_path, crs=gpd.read_file(gpkg_path, layer="XS").crs, new_file=True)
        inst._gpkg_path = gpkg_path
        inst._version = version
        inst._title = title

        inst.contents = inst._content_from_gpkg
        return inst

    @property
    def _content_from_gpkg(
        self,
    ):
        if not hasattr(self, "_gpkg_path"):
            raise ("gpkg_path not provided")
        if not os.path.exists(self._gpkg_path):
            raise FileNotFoundError(f"Could not find the specified gpkg_path {self._gpkg_path}")

        self._check_layers()
        xs_gdf = gpd.read_file(self._gpkg_path, layer="XS", driver="GPKG")
        # headers
        gpkg_data = (
            f"Geom Title={self._title}\nProgram Version={self._version}\nViewing Rectangle={self._bbox(xs_gdf)}\n\n"
        )

        # junction data
        if "Junction" in fiona.listlayers(self._gpkg_path):
            junction_gdf = gpd.read_file(self._gpkg_path, layer="Junction", driver="GPKG")
            junction_gdf["ras_data"].str.cat(sep="\n")

        # river reach data
        gpkg_data += self._river_reach_data_from_gpkg

        return gpkg_data.splitlines()

    @property
    def _river_reach_data_from_gpkg(self):
        river_gdf = gpd.read_file(self._gpkg_path, layer="River", driver="GPKG")
        xs_gdf = gpd.read_file(self._gpkg_path, layer="XS", driver="GPKG")

        if "Structure" in fiona.listlayers(self._gpkg_path):
            structure_gdf = gpd.read_file(self._gpkg_path, layer="Structure", driver="GPKG")
            node_gdf = pd.concat([xs_gdf, structure_gdf]).sort_values(by="river_station", ascending=False)
        else:
            node_gdf = xs_gdf.sort_values(by="river_station", ascending=False)

        data = ""
        for _, row in river_gdf.iterrows():
            centroid = row.geometry.centroid

            coords = row["geometry"].coords

            data += f"River Reach={row['river'].ljust(16)},{row['reach'].ljust(16)}\n"
            data += f"Reach XY= {len(coords)} \n"

            for i, (x, y) in enumerate(coords):
                data += str(x)[:16].rjust(16) + str(y)[:16].rjust(16)
                if i % 2 != 0:
                    data += "\n"

            data += f"\nRch Text X Y={centroid.x},{centroid.y}\nReverse River Text= 0 \n\n"

            # cross section and structures data
            data += node_gdf.loc[node_gdf["river_reach"] == row["river_reach"], "ras_data"].str.cat(sep="\n")

        return data

    def _bbox(self, gdf):
        bounds = gdf.total_bounds
        return f"{bounds[0]} , {bounds[2]} , {bounds[3]} , {bounds[1]} "

    def _check_layers(self):
        layers = set(fiona.listlayers(self._gpkg_path)) & set(SUPPORTED_LAYERS)

        if "XS" not in layers:
            raise NoCrossSectionLayerError(f"Could not find a layer called XS in {self._gpkg_path}")

        if "River" not in layers:
            raise NoRiverLayerError(f"Could not find a layer called River in {self._gpkg_path}")

    @property
    def title(self):
        """Title of the HEC-RAS Geometry file."""
        return search_contents(self.contents, "Geom Title")

    @property
    def version(self):
        """The HEC-RAS version."""
        return search_contents(self.contents, "Program Version", expect_one=False)

    @property
    @check_crs
    def reaches(self) -> dict:
        """A dictionary of the reaches contained in the HEC-RAS geometry file."""
        river_reaches = search_contents(self.contents, "River Reach", expect_one=False)
        reaches = {}
        for river_reach in river_reaches:
            reaches[river_reach] = Reach(self.contents, river_reach, self.crs)
        return reaches

    @property
    @check_crs
    def rivers(self) -> dict:
        """A nested river-reach dictionary of the rivers/reaches contained in the HEC-RAS geometry file."""
        rivers = {}
        for reach in self.reaches.values():
            rivers[reach.river] = {}
            rivers[reach.river].update({reach.reach: reach})
        return rivers

    @property
    @check_crs
    def junctions(self) -> dict:
        """A dictionary of the junctions contained in the HEC-RAS geometry file."""
        juncts = search_contents(self.contents, "Junct Name", expect_one=False)
        junctions = {}
        for junct in juncts:
            junctions[junct] = Junction(self.contents, junct, self.crs)
        return junctions

    @property
    @check_crs
    def cross_sections(self) -> dict:
        """A dictionary of the cross sections contained in the HEC-RAS geometry file."""
        cross_sections = {}
        for reach in self.reaches.values():
            cross_sections.update(reach.cross_sections)

        return cross_sections

    @property
    @check_crs
    def structures(self) -> dict:
        """A dictionary of the structures contained in the HEC-RAS geometry file."""
        structures = {}
        for reach in self.reaches.values():
            structures.update(reach.structures)

        return structures

    @property
    @check_crs
    @add_fid_index
    def reach_gdf(self):
        """A GeodataFrame of the reaches contained in the HEC-RAS geometry file."""
        return pd.concat([reach.gdf for reach in self.reaches.values()], ignore_index=True)

    @property
    @check_crs
    @add_fid_index
    def junction_gdf(self):
        """A GeodataFrame of the junctions contained in the HEC-RAS geometry file."""
        if self.junctions:
            return pd.concat(
                [junction.gdf for junction in self.junctions.values()],
                ignore_index=True,
            )

    @property
    @check_crs
    @add_fid_index
    def xs_gdf(self):
        """Geodataframe of all cross sections in the geometry text file."""
        return pd.concat([xs.gdf for xs in self.cross_sections.values()], ignore_index=True)

    @property
    @check_crs
    @add_fid_index
    def structures_gdf(self):
        """Geodataframe of all structures in the geometry text file."""
        return pd.concat([structure.gdf for structure in self.structures.values()], ignore_index=True)

    @property
    @check_crs
    def n_cross_sections(self):
        """Number of cross sections in the HEC-RAS geometry file."""
        return len(self.cross_sections)

    @property
    @check_crs
    def n_structures(self):
        """Number of structures in the HEC-RAS geometry file."""
        return len(self.structures)

    @property
    @check_crs
    def n_reaches(self):
        """Number of reaches in the HEC-RAS geometry file."""
        return len(self.reaches)

    @property
    @check_crs
    def n_junctions(self):
        """Number of junctions in the HEC-RAS geometry file."""
        return len(self.junctions)

    @property
    @check_crs
    def n_rivers(self):
        """Number of rivers in the HEC-RAS geometry file."""
        return len(self.rivers)

    @check_crs
    def to_gpkg(self, gpkg_path: str):
        """Write the HEC-RAS Geometry file to geopackage."""
        self.xs_gdf.to_file(gpkg_path, driver="GPKG", layer="XS")
        self.reach_gdf.to_file(gpkg_path, driver="GPKG", layer="River")
        if self.junctions:
            self.junction_gdf.to_file(gpkg_path, driver="GPKG", layer="Junction")
        if self.structures:
            self.structures_gdf.to_file(gpkg_path, driver="GPKG", layer="Structure")


class RasFlowText(RasTextFile):
    """Represents a HEC-RAS flow text file."""

    def __init__(self, ras_text_file_path: str, new_file: bool = False):
        super().__init__(ras_text_file_path, new_file)
        if self.file_extension in VALID_UNSTEADY_FLOWS or self.file_extension in VALID_QUASISTEADY_FLOWS:
            raise NotImplementedError("only steady flow (.f**) supported")

        if self.file_extension not in VALID_STEADY_FLOWS:
            raise TypeError(f"Flow extenstion must be one of .f01-.f99, not {self.file_extension}")

    def __repr__(self):
        """Representation of the RasFlowText class."""
        return f"RasFlowText({self._ras_text_file_path})"

    @classmethod
    def from_str(cls, text_string: str, ras_text_file_path: str = ""):
        """Read flow file from string."""
        inst = cls(ras_text_file_path, new_file=True)
        inst.contents = text_string.splitlines()
        return inst

    @property
    def title(self):
        """Title of the flow File."""
        return search_contents(self.contents, "Flow Title")

    @property
    def version(self):
        """Program Version."""
        return search_contents(self.contents, "Program Version")

    @property
    def n_profiles(self):
        """Number of profiles."""
        return len(self.profile_names)

    @property
    def profile_names(self):
        """Profile names."""
        return search_contents(self.contents, "Profile Names").split(",")

    def write_headers(self, title: str, profile_names: list[str]):
        """
        Write headers for flow content.

        Args:
            title (str): title of the flow
            profile_names (list[str]): profile names for the flow

        Returns
        -------
            list (list[str]): lines of the flow content
        """
        lines = [
            f"Flow Title={title}",
            f"Number of Profiles= {len(profile_names)}",
            f"Profile Names={','.join([str(pn) for pn in profile_names])}",
        ]
        return lines

    def write_discharges(self, flows: list, river: str, reach: str, river_station: float):
        """
        Write discharges to flow content.

        Args:
            flows (list): flows to write to the flow content
            river (str): Ras river
            reach (str): Ras reach
            river_station (float): Ras river station
        """
        lines = []
        lines.append(f"River Rch & RM={river},{reach.ljust(16,' ')},{str(river_station).ljust(8,' ')}")
        line = ""
        for i, flow in enumerate(flows):
            line += f"{str(flow).rjust(8,' ')}"
            if (i + 1) % 10 == 0:
                lines.append(line)
                line = ""

        if (i + 1) % 10 != 0:
            lines.append(line)
        return lines

    def write_ds_known_wse(self, ds_wses: list[float], river: str, reach: str):
        """
        Write downstream known water surface elevations to flow content.

        Args:
            ds_wses (list): downstream known water surface elevations
            river (str): Ras river
            reach (str): Ras reach
        """
        lines = []
        for count, wse in enumerate(ds_wses):
            lines.append(f"Boundary for River Rch & Prof#={river},{reach.ljust(16,' ')}, {count+1}")
            lines.append("Up Type= 0 ")
            lines.append("Dn Type= 1 ")
            lines.append(f"Dn Known WS={wse}")

        return lines

    def write_ds_normal_depth(self, number_of_profiles: int, normal_depth: float, river: str, reach: str):
        """
        Write the downstream normal depth boundary condition.

        Args:
            number_of_profiles (int): Number of profiles
            normal_depth (float): Normal depth slope to apply to all profiles
            river (str): Ras river
            reach (str): Ras reach
        """
        lines = []
        for i in range(number_of_profiles):
            lines.append(f"Boundary for River Rch & Prof#={river},{reach.ljust(16,' ')}, {i+1}")
            lines.append("Up Type= 0 ")
            lines.append("Dn Type= 3 ")
            lines.append(f"Dn Slope={normal_depth}")

        return lines

    @property
    def n_flow_change_locations(self):
        """Number of flow change locations."""
        return len(search_contents(self.contents, "River Rch & RM", expect_one=False))

    @property
    def flow_change_locations(self):
        """Retrieve flow change locations."""
        flow_change_locations, locations = [], []
        for location in search_contents(self.contents, "River Rch & RM", expect_one=False):
            # parse river, reach, and river station for the flow change location
            river, reach, rs = location.split(",")
            lines = text_block_from_start_end_str(
                f"River Rch & RM={location}", ["River Rch & RM", "Boundary for River Rch & Prof#"], self.contents
            )
            flows = []

            for line in lines[1:]:

                if "River Rch & RM" in line:
                    break
                for i in range(0, len(line), 8):
                    flows.append(float(line[i : i + 8].lstrip(" ")))
                    if len(flows) == self.n_profiles:
                        flow_change_locations.append(
                            FlowChangeLocation(
                                river,
                                reach.rstrip(" "),
                                float(rs),
                                flows,
                                self.profile_names,
                            )
                        )

                    if len(flow_change_locations) == self.n_flow_change_locations:
                        return flow_change_locations


class RasMap:
    """
    Represents a single RasMapper file.

    Attributes
    ----------
    text_file : Text file repressenting the RAS Mapper file
    contents : a text representation of the files contents
    version : HEC-RAS version for this RAS Mapper file

    """

    def __init__(self, path: str, geom, version: str = 631):
        """
        Represent a single RasMapper file.

        Args:
            path (str): file path to the plan text file
            version (str): HEC-RAS vesion for this RAS Mapper file
        """
        self.text_file = path
        self.version = version

        if os.path.exists(path):
            self.read_contents()
        else:
            self.new_rasmap_content(geom)

    def __repr__(self):
        """Representation of the RasMap class."""
        return f"RasMap({self.path})"

    def read_contents(self):
        """
        Read contents of the file. Searches for the file locally and on s3.

        Raises
        ------
            FileNotFoundError:
        """
        if os.path.exists(self.text_file):
            with open(self.text_file) as f:
                self.contents = f.read()
        else:
            raise FileNotFoundError(f"could not find {self.text_file}")

    def new_rasmap_content(self, geom):
        """Populate the contents with boiler plate contents."""
        if self.version in ["631", "6.31", "6.3.1", "63.1"]:
            self.contents = RASMAP_631.replace("geom_hdf_placeholder", os.path.basename(geom.hdf_file)).replace(
                "geom_name_placeholder", geom.title
            )
        else:
            raise ValueError(f"model version '{self.version}' is not supported")

    def update_crs(self, projection_file: str):
        """
        Add/update the crs file to the RAS Mapper contents.

        Args:
            projection_file (str): path to projeciton file containing the coordinate system (.prj)

        Raises
        ------
            FileNotFoundError:
        """
        directory = os.path.dirname(self.text_file)
        crs_base = os.path.basename(projection_file)

        if crs_base not in os.listdir(directory):
            raise FileNotFoundError(
                f"Expected crs file to be in RAS directory: {directory}. Provided location is: {projection_file}"
            )

        lines = self.contents.splitlines()
        lines.insert(2, rf'  <RASProjectionFilename Filename=".\{crs_base}" />')

        self.contents = "\n".join(lines)

    def add_result_layers(self, plan_short_id: str, profiles: list[str], variable: str):
        """
        Add results layers to RasMap contents. When the RAS plan is ran with "Floodplain Mapping" toggled onthe result layer added here will output rasters.

        Args:
            plan_short_id (str): Plan short id for the output raster(s)
            profiles (list[str]): Profiles for the output raster(s)
            variable (str): Variable to create rasters for. Currently "Depth" is the only supported variable.
        """
        if variable not in ["Depth"]:
            raise NotImplementedError(
                f"Variable {variable} not currently implemented. Currently only Depth is supported."
            )

        lines = []
        for line in self.contents.splitlines():
            if line == "  </Results>":
                for i, profile in enumerate(profiles):
                    lines.append(
                        rf'      <Layer Name="{variable}" Type="RASResultsMap" Checked="True" Filename=".\{plan_short_id}\{variable} ({profile}).vrt">'
                    )
                    lines.append(
                        rf'        <MapParameters MapType="{variable.lower()}" OutputMode="Stored Current Terrain" StoredFilename=".\{plan_short_id}\{variable} ({profile}).vrt" ProfileIndex="{i}" ProfileName="{profile}" />'
                    )
                    lines.append("      </Layer>")
                lines.append("    </Layer>")
            lines.append(line)

        self.contents = "\n".join(lines)

    def add_plan_layer(self, plan_short_id: str, plan_hdf: str, profiles: list[str]):
        """
        Add a plan layer to the results in the RASMap contents.

        Args:
            plan_short_id (str): plan_short_id of the plan
            plan_hdf (str): hdf file for the plan
            profiles (list[str]): profiles for the plan
        """
        lines = []
        for line in self.contents.splitlines():
            if line == "  <Results />":
                lines.append(
                    PLAN.replace("plan_hdf_placeholder", plan_hdf)
                    .replace("plan_name_placeholder", str(plan_short_id))
                    .replace("profile_placeholder", profiles[0])
                )
                continue

            lines.append(line)

        self.contents = "\n".join(lines)

    def add_terrain(self, terrain_name: str, terrain_path: str):
        """Add Terrain to RasMap content."""
        lines = []
        for line in self.contents.splitlines():
            if line == "  <Terrains />":
                lines.append(TERRAIN.replace(TERRAIN_NAME, terrain_name).replace(TERRAIN_PATH, terrain_path))
                continue

            lines.append(line)

        self.contents = "\n".join(lines)

    def write(self):
        """Write Ras Map contents to file."""
        logging.info(f"writing: {self.text_file}")

        with open(self.text_file, "w") as f:
            f.write(self.contents)

        # write backup
        with open(self.text_file + ".backup", "w") as f:
            f.write(self.contents)


# functions
def search_for_ras_crs(search_dir: str):
    """Search for RAS crs."""
    rasmap_files = glob.glob(f"{search_dir}/*.rasmap")
    if rasmap_files:
        rm = rasmap_files[0]
    else:
        raise FileNotFoundError(f"Could not find rasmap file in {search_dir}")

    with open(rm) as f:
        for line in f.readlines():
            if "RASProjectionFilename Filename=" in line:
                relative_path = line.split("=")[-1].split('"')[1].lstrip(".")
                abs_path = f"{search_dir}/{relative_path}"
                if os.path.exists(abs_path):
                    with open(abs_path) as src:
                        crs = src.read()
                        return crs, abs_path
                else:
                    raise FileNotFoundError(f"Could not find crs file in {search_dir}")


def get_new_extension_number(dict_of_ras_subclasses: dict) -> str:
    """
    Determine the next numeric extension that should be used when creating a new plan, flow, or geom; e.g., if you are adding a new plan and .p01, and .p02 already exists then the new planwill have a .p03 extension.

    Args:
        dict_of_ras_subclasses (dict): A dictionary containing plan/geom/flow titles as keys
            and objects plan/geom/flow as values.

    Returns
    -------
        new file extension (str): The new file exension.
    """
    extension_number = []
    if not dict_of_ras_subclasses:
        return "01"
    for val in dict_of_ras_subclasses.values():
        extension_number.append(int(val.file_extension[2:]))

    return f"{(max(extension_number)+1):02d}"


def create_terrain(
    src_terrain_filepaths: list[str],
    projection_file: str,
    dst_terrain_filepath: str,
    vertical_units: str = "Feet",
    version: str = "631",
) -> str:
    r"""
    Use the crs file and a list of terrain file paths to make the RAS terrain HDF file. Default location is {model_directory}\Terrain\Terrain.hdf.

    Returns the full path to the local directory containing the output files.
    """
    if vertical_units not in ["Feet", "Meters"]:
        raise ValueError(f"vertical_units must be either 'Feet' or 'Meters'; got: '{vertical_units}'")

    missing_files = [x for x in src_terrain_filepaths if not os.path.exists(x)]

    if missing_files:
        raise FileNotFoundError(str(missing_files))

    terrain_exe = get_terrain_exe_path(version)
    if not os.path.isfile(terrain_exe):
        raise FileNotFoundError(terrain_exe)

    exe_parent_dir = os.path.split(terrain_exe)[0]
    # TODO: Add documentation for the following to understand
    # what files are created and where they are stored
    subproc_args = [
        terrain_exe,
        "CreateTerrain",
        f"units={vertical_units}",  # vertical units
        "stitch=true",
        f"prj={projection_file}",
        f"out={dst_terrain_filepath}",
    ]
    # add list of input rasters from which to build the Terrain
    subproc_args.extend([os.path.abspath(p) for p in src_terrain_filepaths])
    logging.debug(f"Running the following args, from {exe_parent_dir}:" + "\n  ".join([""] + subproc_args))
    subprocess.check_call(subproc_args, cwd=exe_parent_dir, stdout=subprocess.DEVNULL)
    return f"Terrain written to {dst_terrain_filepath}"

    # TODO this recompression does work but RAS does not accept the recompressed tif for unknown reason...
    # # compress the output tif(s) that RasProcess.exe created (otherwise could be 1+ GB at HUC12 size)
    # layer_name = os.path.splitext(hdf_filename)[0]  # hdf file without extension
    # tif_pattern = rf"^{layer_name}\..+\.tif$"
    # output_tifs = [
    #     os.path.join(terrain_dir_fp, fn) for fn in os.listdir(terrain_dir_fp) if re.fullmatch(tif_pattern, fn)
    # ]
    # for tif in output_tifs:
    #     utils.recompress_tif(tif)
    #     utils.build_tif_overviews(tif)



"""Conflation classes and functions."""

import json
import logging
import os
import sqlite3
from collections import OrderedDict
from typing import List, Tuple

import fiona
import geopandas as gpd
import numpy as np
import pandas as pd
import pyproj
from fiona.errors import DriverError
from shapely.geometry import LineString, MultiLineString, MultiPoint, Point, Polygon, box
from shapely.ops import linemerge, nearest_points, transform

from ripple1d.consts import METERS_PER_FOOT
from ripple1d.utils.ripple_utils import xs_concave_hull

HIGH_FLOW_FACTOR = 1.2

NWM_CRS = """PROJCRS["USA_Contiguous_Albers_Equal_Area_Conic_USGS_version",
    BASEGEOGCRS["NAD83",DATUM["North American Datum 1983",
    ELLIPSOID["GRS 1980",6378137,298.257222101004,LENGTHUNIT["metre",1]]],
    PRIMEM["Greenwich",0,ANGLEUNIT["degree",0.0174532925199433]],
    ID["EPSG",4269]],CONVERSION["unnamed",METHOD["Albers Equal Area",
    ID["EPSG",9822]],PARAMETER["Latitude of 1st standard parallel"
    ,29.5,ANGLEUNIT["degree",0.0174532925199433],ID["EPSG",8823]]
    ,PARAMETER["Latitude of 2nd standard parallel",45.5,
    ANGLEUNIT["degree",0.0174532925199433],ID["EPSG",8824]],
    PARAMETER["Latitude of false origin",23,ANGLEUNIT["degree",0.0174532925199433],
    ID["EPSG",8821]],PARAMETER["Longitude of false origin",-96,
    ANGLEUNIT["degree",0.0174532925199433],ID["EPSG",8822]],
    PARAMETER["Easting at false origin",0,LENGTHUNIT["metre",1],
    ID["EPSG",8826]],PARAMETER["Northing at false origin",0,
    LENGTHUNIT["metre",1],ID["EPSG",8827]]],CS[Cartesian,2],
    AXIS["easting",east,ORDER[1],LENGTHUNIT["metre",1,ID["EPSG",9001]]],
    AXIS["northing",north,ORDER[2],LENGTHUNIT["metre",1,ID["EPSG",9001]]]]"""


def ensure_geometry_column(gdf: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
    """Confirm that there exists a geometry column in the GeoDataFrame."""
    if "geom" in gdf.columns:
        return gdf
    elif "geometry" in gdf.columns:
        gdf.rename(columns={"geometry": "geom"}, inplace=True)
        gdf.set_geometry("geom")
        return gdf
    else:
        raise KeyError(f"Expecting a `geom` or `geometry` column, did not find in {gdf.columns}")


class RasFimConflater:
    """
    Conflate NWM and RAS data for a single river reach.

    Args:
        nwm_parquet (str): Path to the NWM Parquet file converted to parquet from:
            s3://noaa-nws-owp-fim/rasfim/inputs/X-National_Datasets/nwm_flows.gpkg
        ras_gpkg (str): Path to the RAS GeoPackage
        load_data (bool, optional): Load the data on initialization. Defaults to True.

    Raises
    ------
        ValueError: Required layer not found in the GeoPackage
        DriverError: Unable to read the GeoPackage
    """

    def __init__(
        self, nwm_pq: str, source_model_directory: str, load_data: bool = True, output_concave_hull_path: str = None
    ):
        self.nwm_pq = nwm_pq
        self.source_model_directory = source_model_directory
        self.ras_model_name = os.path.basename(source_model_directory)
        self.ras_gpkg = os.path.join(source_model_directory,f"{self.ras_model_name}.gpkg")

        self.output_concave_hull_path = output_concave_hull_path
        self._nwm_reaches = None

        self._ras_centerlines = None
        self._ras_xs = None
        self._ras_structures = None
        self._ras_junctions = None
        self._ras_metadata=None
        self._common_crs = None
        self._xs_hulls = None
        self.__data_loaded = False
        if load_data:
            self._common_crs = NWM_CRS
            self.load_data()
            self.__data_loaded = True
            

    def __repr__(self):
        """Return the string representation of the object."""
        return f"RasFimConflater(nwm_pq={self.nwm_pq}, ras_gpkg={self.ras_gpkg})"


    @property
    def stac_api(self):
        """The stac_api for the HEC-RAS Model."""
        if self.ras_metadata:
            if "stac_api" in self.ras_metadata.keys():
                return self.ras_metadata["stac_api"]
    
    @property
    def stac_collection_id(self):
        """The stac_collection_id for the HEC-RAS Model."""
        if self.ras_metadata:
            if "stac_collection_id" in self.ras_metadata.keys():
                return self.ras_metadata["stac_collection_id"]
        
    
    @property
    def stac_item_id(self):
        """The stac_item_id for the HEC-RAS Model."""
        if self.ras_metadata:
            if "stac_item_id" in self.ras_metadata.keys():
                return self.ras_metadata["stac_item_id"]

    @property
    def primary_geom_file(self):
        """The primary geometry file for the HEC-RAS Model."""
        if self.ras_metadata:
            return self.ras_metadata["primary_geom_file"]

    @property
    def primary_flow_file(self):
        """The primary flow file for the HEC-RAS Model."""
        if self.ras_metadata:
            return self.ras_metadata["primary_flow_file"]
    
    @property
    def primary_plan_file(self):
        """The primary plan file for the HEC-RAS Model."""
        if self.ras_metadata:
            return self.ras_metadata["primary_plan_file"]

    @property
    def ras_project_file(self):
        """The source HEC-RAS project file."""
        if self.ras_metadata:
            return self.ras_metadata["ras_project_file"] 

    # @property
    # def xs_length_units(self):
    #     """Length units of the source HEC-RAS model."""
    #     if self.ras_metadata["units"] != "English":
    #         raise NotImplementedError(
    #             f"HEC-RAS units are {self.ras_metadata['units']}. Only 'English' units are supported at this time."
    #         )
    #     elif self.ras_metadata["units"] == "English":
    #         ras_xs=self.ras_xs
    #         ras_xs["r"]=ras_xs.apply(lambda row: self.populate_r_station(row),axis=1)
    #         if ras_xs["r"].mean()>.9 and ras_xs["r"].mean()<1.1:
    #             return "ft"
    #         elif ras_xs["r"].mean()*METERS_PER_FOOT>.9 and ras_xs["r"].mean()*METERS_PER_FOOT<1.1:
    #             raise ValueError(f"HEC-RAS units specified as English but cross section r values indicate meters")
    #         elif ras_xs["r"].mean()/5280>.9 and ras_xs["r"].mean()/5280<1.1:
    #             return "miles"
    #         else:
    #             raise ValueError(f"Unable to determine cross section length units from cross section r values")

    # @property
    # def river_station_units(self):
    #     """Station units of the source HEC-RAS model."""
    #     if self.ras_metadata["units"] != "English":
    #         raise NotImplementedError(
    #             f"HEC-RAS units are {self.ras_metadata['units']}. Only 'English' units are supported at this time."
    #         )
    #     elif self.ras_metadata["units"] == "English":
    #         ras_xs=self.ras_xs
    #         ras_xs["intersection_point"]=ras_xs.apply(lambda row: self.ras_centerlines(row.geometry).intersection(row.geometry),axis=1)
    #         ras_xs["computed_river_station"]=ras_xs.apply(lambda row: self.ras_centerlines.project(row["intersection_point"])*METERS_PER_FOOT,axis=1)
    #         ras_xs["computed_reach_length"]=ras_xs["computed_river_station"].diff()

    #         ras_xs["reach_lengths_from_original_river_station"]=ras_xs["river_station"].diff()
    #         ras_xs["reach_length_ratio"]=ras_xs["computed_reach_length"]/ras_xs["reach_lengths_from_original_river_station"]
    #         if ras_xs["reach_length_ratio"].mean()>.9 and ras_xs["reach_length_ratio"].mean()<1.1:
    #             return "ft"
    #         elif ras_xs["reach_length_ratio"].mean()*METERS_PER_FOOT>.9 and ras_xs["reach_length_ratio"].mean()*METERS_PER_FOOT<1.1:
    #             raise ValueError(f"HEC-RAS units specified as English but reach length r values indicate meters")
    #         elif ras_xs["reach_length_ratio"].mean()/5280>.9 and ras_xs["reach_length_ratio"].mean()/5280<1.1:
    #             return "miles"
    #         else:
    #             raise ValueError(f"Unable to determine reach length units from reach length r values")
            
        
    # @property
    # def flow_units(self):
    #     """Flow units of the source HEC-RAS model."""
    #     if self.gpkg_metadata["units"] != "English":
    #         raise NotImplementedError(
    #             f"HEC-RAS units are {self.ras_metadata['units']}. Only 'English' units are supported at this time."
    #         )
    #     elif self.ras_metadata["units"] == "English":
    #         return "cfs"

    def populate_r_station(self, row: pd.Series,assume_ft:bool=True) -> str:
        """Populate the r value for a cross section. The r value is the ratio of the station to actual cross section length."""
        #TODO check if this is the correct way to calculate r
        df = pd.DataFrame(row["station_elevation_points"], index=["elevation"]).T
        return df.index.max()/(row.geometry.length/METERS_PER_FOOT)

    @property
    def _gpkg_metadata(self):
        """Metadata from gpkg."""
        with sqlite3.connect(self.ras_gpkg) as conn:
            cur = conn.cursor()
            cur.execute("select * from metadata")
            return dict(cur.fetchall())

    def determine_station_order(self, xs_gdf: gpd.GeoDataFrame,reach:LineString):
        """Detemine the order based on stationing of the cross sections along the reach."""
        rs=[]
        for _, xs in xs_gdf.iterrows():
            geom=reach.intersection(xs.geometry)
            try:
                point=validate_point(geom)
                rs.append(reach.project(point))
            except TypeError as e:
                rs.append(rs[-1])
            
        xs_gdf["rs"]=rs
        return xs_gdf.sort_values(by="rs",ignore_index=True)

    def add_hull(self, xs_gdf: gpd.GeoDataFrame,reach:LineString):
        """Add the concave hull to the GeoDataFrame."""
        if len(xs_gdf) > 1:
            xs_gdf=self.determine_station_order(xs_gdf,reach)
            hull = xs_concave_hull(xs_gdf)
            if self._xs_hulls is None:
                self._xs_hulls = hull
            else:
                self._xs_hulls = pd.concat([self._xs_hulls, hull])

    def write_hulls(self):
        """Write the hulls to a GeoPackage."""
        if self._xs_hulls is not None:
            self._xs_hulls.to_file(self.output_concave_hull_path, driver="GPKG", layer="xs_concave_hulls")

    def set_ras_gpkg(self, ras_gpkg, load_data=True):
        """Set the RAS GeoPackage and optionally load the data."""
        self.ras_gpkg = ras_gpkg
        if load_data:
            self.load_gpkg(self.ras_gpkg)

    def load_gpkg(self, gpkg: str):
        """Load the RAS data from the GeoPackage."""
        layers = fiona.listlayers(gpkg)
        if "River" in layers:
            self._ras_centerlines = gpd.read_file(self.ras_gpkg, layer="River")
        if "XS" in layers:
            xs=gpd.read_file(self.ras_gpkg, layer="XS")
            self._ras_xs = xs[xs.intersects(self._ras_centerlines.union_all())]
        if "Junction" in layers:
            self._ras_junctions = gpd.read_file(self.ras_gpkg, layer="Junction")
        if "Structure" in layers:
            structures=gpd.read_file(self.ras_gpkg, layer="Structure")
            self._ras_structures = structures[structures.intersects(self._ras_centerlines.union_all())]
        if "metadata" in layers:
            self._ras_metadata=self._gpkg_metadata

    def load_pq(self, nwm_pq: str):
        """Load the NWM data from the Parquet file."""
        try:
            nwm_reaches = gpd.read_parquet(nwm_pq,bbox=self._ras_xs.to_crs(self.common_crs).total_bounds)
            nwm_reaches = nwm_reaches.rename(columns={"geom": "geometry"})
            self._nwm_reaches = nwm_reaches.set_geometry("geometry")
        except Exception as e:
            if type(e) == DriverError:
                raise DriverError(f"Unable to read {nwm_pq}")
            else:
                raise (e)

    def load_data(self):
        """Load the NWM and RAS data from the GeoPackages."""
        self.load_gpkg(self.ras_gpkg)
        self.load_pq(self.nwm_pq)

    def ensure_data_loaded(func):
        """Ensure that the data is loaded before accessing the properties Decorator."""

        def wrapper(self, *args, **kwargs):
            if not self.__data_loaded:
                raise Exception("Data not loaded")
            return func(self, *args, **kwargs)

        return wrapper

    @property
    def common_crs(self):
        """Return the common CRS for the NWM and RAS data."""
        return self._common_crs

    @property
    @ensure_data_loaded
    def ras_centerlines(self) -> gpd.GeoDataFrame:
        """RAS centerlines."""
        return self._ras_centerlines.to_crs(self.common_crs)

    @property
    @ensure_data_loaded
    def ras_river_reach_names(self) -> List[str]:
        """Return the unique river reach names in the RAS data."""
        return self.ras_centerlines["river_reach"].unique().tolist()

    @property
    @ensure_data_loaded
    def ras_xs(self) -> gpd.GeoDataFrame:
        """RAS cross sections."""
        return self._ras_xs.to_crs(self.common_crs)

    @property
    @ensure_data_loaded
    def ras_structures(self) -> gpd.GeoDataFrame:
        """RAS structures."""
        logging.info("RAS structures")
        try:
            return self._ras_structures.to_crs(self.common_crs)
        except AttributeError:
            return None
        
    @property
    @ensure_data_loaded
    def ras_junctions(self) -> gpd.GeoDataFrame:
        """RAS junctions."""
        try:
            return self._ras_junctions.to_crs(self.common_crs)
        except AttributeError:
            return None

    @property
    def ras_metadata(self) -> Polygon:
        """RAS metadata."""
        try:
            return self._ras_metadata
        except AttributeError:
            return None

    @property
    def ras_xs_bbox(self) -> Polygon:
        """Return the bounding box for the RAS cross sections."""
        return self.get_projected_bbox(self.ras_xs)

    def ras_xs_concave_hull(self, river_reach_name: str = None) -> Polygon:
        """Return the concave hull of the cross sections."""
        if river_reach_name is None:
            return xs_concave_hull(self.ras_xs,self.ras_junctions)
        else:
            return xs_concave_hull(self.ras_xs[self.ras_xs["river_reach"] == river_reach_name])

    def ras_xs_convex_hull(self, river_reach_name: str = None):
        """Return the convex hull of the cross sections."""
        if river_reach_name:
            polygon = self.ras_xs["geometry"].union_all().convex_hull

        else:
            polygon = self.ras_xs["geometry"].union_all().convex_hull

        return gpd.GeoDataFrame({"geometry": [polygon]}, geometry="geometry", crs=self.ras_xs.crs)

    @property
    def ras_banks(self):
        """Return the banks of the RAS cross sections."""
        raise NotImplementedError

    def get_projected_bbox(self, gdf: gpd.GeoDataFrame) -> Polygon:
        """Return the bounding box for the GeoDataFrame in the common CRS."""
        assert gdf.crs, "GeoDataFrame must have a CRS"
        project = pyproj.Transformer.from_crs(gdf.crs, NWM_CRS, always_xy=True).transform
        return transform(project, box(*tuple(gdf.total_bounds)))

    @property
    @ensure_data_loaded
    def nwm_reaches(self) -> gpd.GeoDataFrame:
        """NWM reaches."""
        return self._nwm_reaches

    def local_nwm_reaches(self, river_reach_name: str = None) -> gpd.GeoDataFrame:
        """NWM reaches that intersect the RAS cross sections."""
        if river_reach_name:

            return self.nwm_reaches[
                self.nwm_reaches.intersects(self.ras_xs_convex_hull(river_reach_name)["geometry"].iloc[0])
            ]
        else:
            return self.nwm_reaches[self.nwm_reaches.intersects(self.ras_xs_convex_hull()["geometry"].iloc[0])]

    @property
    def local_gages(self) -> dict:
        """Local gages for the NWM reaches."""
        local_reaches = self.local_nwm_reaches()
        gages = local_reaches["gages"]
        reach_ids = local_reaches["ID"]
        return {reach_id: gage for reach_id, gage in zip(reach_ids, gages) if pd.notna(gage) and gage.strip()}

    def local_lakes(self):
        """Local lakes for the NWM reaches."""
        raise NotImplementedError

    def check_centerline(func):
        """
        Ensure that ras_centerline is available and that the centerline is specified.

        Helper function.

        If the centerline is not specified, the first centerline is used, with a check that there is only one centerline.
        """

        def wrapper(self, *args, **kwargs):
            assert self.__data_loaded, "Data not loaded"

            river_reach_name = kwargs.get("river_reach_name", None)

            if river_reach_name:
                logging.debug(f"check_centerline river_reach_name: {river_reach_name}")
                centerline = self.ras_centerline_by_river_reach_name(river_reach_name)
            else:
                if self.ras_centerlines.shape[0] == 1:
                    centerline = self.ras_centerlines.geometry.iloc[0]
                else:
                    raise ValueError("Multiple centerlines found, please specify river_reach_name")
            kwargs["centerline"] = centerline
            return func(self, *args, **kwargs)

        return wrapper

    @check_centerline
    def ras_start_end_points(self, river_reach_name: str = None, centerline=None,clip_to_xs=False) -> Tuple[Point, Point]:
        """River_reach_name used by the decorator to get the centerline."""
        if river_reach_name:
            centerline = self.ras_centerline_by_river_reach_name(river_reach_name,clip_to_xs)
        return endpoints_from_multiline(centerline)

    def ras_centerline_by_river_reach_name(self, river_reach_name: str,clip_to_xs=False) -> LineString:
        """Return the centerline for the specified river reach."""
        if clip_to_xs:
            return self.ras_centerlines[self.ras_centerlines["river_reach"] == river_reach_name].geometry.iloc[0].intersection(self.ras_xs_concave_hull(river_reach_name).geometry.iloc[0].buffer(1))
        else:
            return self.ras_centerlines[self.ras_centerlines["river_reach"] == river_reach_name].geometry.iloc[0]

    def xs_by_river_reach_name(self, river_reach_name: str) -> gpd.GeoDataFrame:
        """Return the cross sections for the specified river reach."""
        return self.ras_xs[self.ras_xs["river_reach"] == river_reach_name]

    def river_reach_name_by_xs(self, xs_id: str) -> Tuple[str, str]:
        """Return the river and reach names for the specified cross section."""
        data = self.ras_xs[self.ras_xs["ID"] == int(xs_id)]
        if data.empty:
            raise ValueError(f"XS ID {xs_id} not found in ras_xs")
        elif data.shape[0] != 1:
            raise ValueError(f"Multiple XS found with ID = {xs_id}")
        else:
            return data.iloc[0]["river"], data.iloc[0]["reach"]


# general geospatial functions
def endpoints_from_multiline(mline: MultiLineString) -> Tuple[Point, Point]:
    """Return the start and end points of a MultiLineString."""
    if isinstance(mline, MultiLineString):
        # TODO: Ensure the centerlines in the GPKG are always the same type (i.e. MultiLineString)
        merged_line = linemerge(mline)
    else:
        merged_line = mline
    start_point, end_point = merged_line.coords[0], merged_line.coords[-1]
    return Point(start_point), Point(end_point)


def nearest_line_to_point(lines: gpd.GeoDataFrame, point: Point, column_id: str = "ID") -> int:
    """Return the ID of the line closest to the point."""
    if not column_id in lines.columns:
        raise ValueError(f"required `ID` column not found in GeoDataFrame with columns: {lines.columns}")

    start_reach_distance = 1e9
    start_reach_id = None
    for _, row in lines.iterrows():
        start_distance = row["geometry"].distance(point)
        if start_distance < start_reach_distance:
            start_reach_distance = start_distance
            start_reach_id = row[column_id]

    if start_reach_distance > 1e8:
        raise ValueError(
            f"Unable to associate reach with point, minimum distance for {start_reach_id} : {start_reach_distance}"
        )
    return start_reach_id


def convert_linestring_to_points(
    linestring: LineString, crs: pyproj.crs.crs.CRS, point_spacing: int = 5
) -> gpd.GeoDataFrame:
    """Convert a LineString to a GeoDataFrame of Points."""
    num_points = int(round(linestring.length / point_spacing))

    if num_points == 0:
        num_points = 1

    points = [linestring.interpolate(distance) for distance in np.linspace(0, linestring.length, num_points)]

    return gpd.GeoDataFrame(
        geometry=points,
        crs=crs,
    )


def cacl_avg_nearest_points(reference_gdf: gpd.GeoDataFrame, compare_points_gdf: gpd.GeoDataFrame) -> float:
    """Calculate the average distance between the reference points and the nearest points in the comparison GeoDataFrame."""
    multipoint = compare_points_gdf.geometry.union_all()
    reference_gdf["nearest_distance"] = reference_gdf.geometry.apply(
        lambda point: point.distance(nearest_points(point, multipoint)[1])
    )
    return reference_gdf["nearest_distance"].mean()


def count_intersecting_lines(ras_xs: gpd.GeoDataFrame, nwm_reaches: gpd.GeoDataFrame) -> int:
    """Return the number of intersecting lines between the RAS cross sections and the NWM reaches."""
    if ras_xs.crs != nwm_reaches.crs:
        ras_xs = ras_xs.to_crs(nwm_reaches.crs)
    join_gdf = gpd.sjoin(ras_xs, nwm_reaches, predicate="intersects")
    return join_gdf


def filter_gdf(gdf: gpd.GeoDataFrame, column_values: list, column_name: str = "ID") -> gpd.GeoDataFrame:
    """Filter a GeoDataFrame based on the values in a column."""
    return gdf[~gdf[column_name].isin(column_values)]


# analytical functions
def walk_network(gdf: gpd.GeoDataFrame, start_id: int, stop_id: int) -> List[int]:
    """Walk the network from the start ID to the stop ID."""
    current_id = start_id
    ids = [current_id]

    while current_id != stop_id:
        result = gdf.query(f"ID == {current_id}")

        if result.empty:
            logging.error(f"No row found with ID = {current_id}")
            break

        to_value = result.iloc[0]["to_id"]
        ids.append(int(to_value))
        current_id = to_value
        if current_id == stop_id:
            break

    return ids


# def calculate_conflation_metrics(
#     rfc: RasFimConflater,
#     candidate_reaches: gpd.GeoDataFrame,
#     xs_group: gpd.GeoDataFrame,
#     ras_points: gpd.GeoDataFrame,
# ) -> dict:
#     """Calculate the conflation metrics for the candidate reaches."""
#     next_round_candidates = []
#     xs_hits_ids = []
#     total_hits = 0
#     for i in candidate_reaches.index:
#         candidate_reach_points = convert_linestring_to_points(candidate_reaches.loc[i].geometry, crs=rfc.common_crs)
#         # TODO: Evaluate this constant.
#         if cacl_avg_nearest_points(candidate_reach_points, ras_points) < 10000:
#             next_round_candidates.append(candidate_reaches.loc[i]["ID"])
#             gdftmp = gpd.GeoDataFrame(geometry=[candidate_reaches.loc[i].geometry], crs=rfc.nwm_reaches.crs)
#             xs_hits = count_intersecting_lines(xs_group, gdftmp)

#             total_hits += xs_hits.shape[0]
#             xs_hits_ids.extend(xs_hits.ID.tolist())

#             logging.debug(f"conflation: {total_hits} xs hits out of {xs_group.shape[0]}")

#     dangling_xs = filter_gdf(xs_group, xs_hits_ids)

#     dangling_xs_interesects = gpd.sjoin(dangling_xs, rfc.nwm_reaches, predicate="intersects")

#     conflation_score = round(total_hits / xs_group.shape[0], 2)

#     if conflation_score == 1:
#         conlfation_notes = "Probable Conflation, no dangling xs"
#         manual_check_required = False

#     # elif dangling_xs_interesects.shape[0] == 0:
#     #     conlfation_notes = f"Probable Conflation..."
#     #     manual_check_required = False

#     elif conflation_score >= 0.95:
#         conlfation_notes = f"Probable Conflation: partial nwm reach coverage with {dangling_xs.shape[0]}/{xs_group.shape[0]} dangling xs"
#         manual_check_required = False

#     elif conflation_score >= 0.25:
#         conlfation_notes = f"Possible Conflation: partial nwm reach coverage with {dangling_xs.shape[0]}/{xs_group.shape[0]} dangling xs"
#         manual_check_required = True

#     elif conflation_score < 0.25:
#         conlfation_notes = f"Unable to conflate: potential disconnected reaches with {dangling_xs.shape[0]}/{xs_group.shape[0]} dangling xs"
#         manual_check_required = True

#     elif conflation_score > 1:
#         conlfation_notes = f"Unable to conflate: potential diverging reaches with {dangling_xs.shape[0]}/{xs_group.shape[0]} dangling xs"
#         manual_check_required = True

#     else:
#         conlfation_notes = "Unknown error"
#         manual_check_required = True

#     # Convert next_round_candidates from int64 to serialize
#     conlfation_metrics = {
#         # "fim_reaches": [int(c) for c in next_round_candidates],
#         "conflation_score": round(total_hits / xs_group.shape[0], 2),
#         "conlfation_notes": conlfation_notes,
#         "manual_check_required": manual_check_required,
#     }
#     return conlfation_metrics


def ras_xs_geometry_data(rfc: RasFimConflater, xs_id: str) -> dict:
    """Return the geometry data (max/min xs elevation) for the specified cross section."""
    # TODO: Need to verify units in the RAS data
    xs = rfc.ras_xs[rfc.ras_xs["ID"] == xs_id]
    if xs.shape[0] > 1:
        raise ValueError(f"Multiple XS found with ID = {xs_id}")

    elif xs.shape[0] == 0:
        raise ValueError(f"No XS found with ID = {xs_id}")
    else:
        xs = xs.iloc[0]

    xs.rename({"thalweg": "min_elevation", "xs_max_elevation": "max_elevation", "river_station": "xs_id"}, inplace=True)
    xs_data = xs[["river", "reach", "min_elevation", "max_elevation", "xs_id"]].to_dict()

    return xs[["river", "reach", "min_elevation", "max_elevation", "xs_id"]].to_dict()


def get_us_most_xs_from_junction(rfc, us_river, us_reach):
    """Search for river reaches at junctions and return xs closest ds reach at junciton."""
    if rfc.ras_junctions is None:
        raise ValueError("No junctions found in the HEC-RAS model.")
    ds_rivers, ds_reaches = None, None
    for row in rfc.ras_junctions.itertuples():
        us_rivers = row.us_rivers.split(",")
        us_reaches = row.us_reaches.split(",")
        for river, reach in zip(us_rivers, us_reaches):

            if river == us_river and reach == us_reach:
                ds_rivers = row.ds_rivers.split(",")
                ds_reaches = row.ds_reaches.split(",")
                break

    if not ds_rivers:
        raise ValueError(f"No downstream rivers found for {us_river}")
    elif len(ds_rivers) == 0:
        raise ValueError(f"No downstream rivers found for {us_river}")
    elif len(ds_rivers) > 1:
        raise NotImplementedError(f"Multiple downstream rivers found for {us_river}")
    else:
        ds_river = ds_rivers[0]

    if not ds_reaches:
        raise ValueError(f"No downstream reaches found for {us_reach}")
    elif len(ds_reaches) == 0:
        raise ValueError(f"No downstream reaches found for {us_reach}")
    elif len(ds_reaches) > 1:
        raise NotImplementedError(f"Multiple downstream reaches found for {us_reach}")
    else:
        ds_reach = ds_reaches[0]

    ds_river_reach_gdf = rfc.ras_xs.loc[(rfc.ras_xs["river"] == ds_river) & (rfc.ras_xs["reach"] == ds_reach), :]
    max_rs = ds_river_reach_gdf.loc[:, "river_station"].max()
    ds_xs_id = int(ds_river_reach_gdf.loc[ds_river_reach_gdf["river_station"] == max_rs, "ID"].iloc[0])
    return ds_xs_id


def validate_point(geom):
    """Validate that point is of type Point. If Multipoint or Linestring create point from first coordinate pair."""
    if isinstance(geom, Point):
        return geom
    elif isinstance(geom, MultiPoint):
        return geom.geoms[0]
    elif isinstance(geom, LineString) and list(geom.coords):
        return Point(geom.coords[0])
    else:
        raise TypeError(f"expected point at xs-river intersection got: {type(geom)}")


def check_xs_direction(cross_sections: gpd.GeoDataFrame, reach: LineString):
    """Return only cross sections that are drawn right to left looking downstream."""
    ids = []
    for _, xs in cross_sections.iterrows():
        try:
            point = reach.intersection(xs["geometry"])
            point = validate_point(point)
            xs_rs = reach.project(point)

            offset = xs.geometry.offset_curve(-1)
            point = reach.intersection(offset)
            point = validate_point(point)

            offset_rs = reach.project(point)
            if xs_rs > offset_rs:
                ids.append(xs["ID"])
        except TypeError as e:
            logging.warning(f"could not validate xs-river intersection for: {xs["river"]} {xs['reach']} {xs['river_station']}")
            continue
    return cross_sections.loc[cross_sections["ID"].isin(ids)]


def map_reach_xs(rfc: RasFimConflater, reach: MultiLineString) -> dict:
    """
    Map the upstream and downstream cross sections for the nwm reach.

    TODO: This function needs helper functions to extract the datasets from the GeoDataFrames.
    """
    # get the xs that intersect the nwm reach
    intersected_xs = rfc.ras_xs[rfc.ras_xs.intersects(reach.geometry)]
    intersected_xs = check_xs_direction(intersected_xs, reach.geometry)
    has_junctions = rfc.ras_junctions is not None

    if intersected_xs.empty:
        return {"eclipsed":True}

    # get start and end points of the nwm reach
    start, end = endpoints_from_multiline(reach.geometry)

    # Begin with us_xs data
    us_xs = nearest_line_to_point(intersected_xs, start, column_id="ID")

    # Initialize us_xs data with min /max elevation, then build the dict with added info
    us_data = ras_xs_geometry_data(rfc, us_xs)

    # Add downstream xs data
    ds_xs = nearest_line_to_point(intersected_xs, end, column_id="ID")

    # get infor for the current ds_xs
    ras_river = rfc.ras_xs.loc[rfc.ras_xs["ID"] == ds_xs, "river"].iloc[0]
    ras_reach = rfc.ras_xs.loc[rfc.ras_xs["ID"] == ds_xs, "reach"].iloc[0]
    river_station = rfc.ras_xs.loc[rfc.ras_xs["ID"] == ds_xs, "river_station"].iloc[0]
    river_reach_gdf = rfc.ras_xs.loc[(rfc.ras_xs["river"] == ras_river) & (rfc.ras_xs["reach"] == ras_reach), :]

    # check if this is the most downstream xs on this ras reach
    if river_reach_gdf["river_station"].min() == river_station:
        try:
            ds_xs = get_us_most_xs_from_junction(rfc, ras_river, ras_reach)
            ds_data = ras_xs_geometry_data(rfc, ds_xs)
            logging.debug(
                f"Conflating to junction for reach {reach.ID}: {us_data['river']} {us_data['reach']} to {ds_data['river']} {ds_data['reach']}"
            )
        except ValueError as e:
            ds_data = ras_xs_geometry_data(rfc, ds_xs)
            logging.warning(f"No downstream XS's: {e}")

    # if it is not the most downstream xs on this ras reach then compute the next downstream river station and determine the id
    else:
        ds_xs_gdf = river_reach_gdf.loc[river_reach_gdf["river_station"] < river_station].reset_index(drop=True)
        if len(ds_xs_gdf) == 1:
            index = 0
        else:
            index = ds_xs_gdf["river_station"].idxmax()
        ds_xs = ds_xs_gdf["ID"].iloc[index]
        ds_data = ras_xs_geometry_data(rfc, ds_xs)

    # add xs concave hull
    if rfc.output_concave_hull_path:
        xs_gdf = pd.concat([intersected_xs, rfc.ras_xs[rfc.ras_xs["ID"] == ds_xs]], ignore_index=True)
        rfc.add_hull(xs_gdf,reach.geometry)
    
    if us_data==ds_data:
        return {"eclipsed":True}

    return {
        "us_xs": us_data,
        "ds_xs": ds_data,
        "eclipsed":False
    }


def ras_reaches_metadata(rfc: RasFimConflater, candidate_reaches: gpd.GeoDataFrame):
    """Return the metadata for the RAS reaches."""
    reach_metadata = OrderedDict()
    hulls = []
    for reach in candidate_reaches.itertuples():
        # logging.debug(f"REACH: {reach.ID}")

        # get the xs data for the reach
        ras_xs_data = map_reach_xs(rfc, reach)

        reach_metadata[reach.ID] = ras_xs_data

    for k in reach_metadata.keys():
        flow_data = rfc.nwm_reaches[rfc.nwm_reaches["ID"] == k].iloc[0]
        if isinstance(flow_data["high_flow_threshold"], float):
            reach_metadata[k]["low_flow"] = int(round(flow_data["high_flow_threshold"], 2) * HIGH_FLOW_FACTOR)
        else:
            reach_metadata[k]["low_flow"] = -9999
            logging.warning(f"No low flow data for {k}")

        try:
            high_flow = float(flow_data["f100year"])
            reach_metadata[k]["high_flow"] = int(round(high_flow, 2))
        except:
            logging.warning(f"No high flow data for {k}")
            reach_metadata[k]["high_flow"] = -9999
        try:
            reach_metadata[k]["network_to_id"] = str(flow_data["to_id"])
        except:
            logging.warning(f"No to_id data for {k}")
            reach_metadata[k]["network_to_id"] = "-9999"

        if k in rfc.local_gages.keys():
            gage_id = rfc.local_gages[k].replace(" ", "")
            reach_metadata[k]["gage"] = gage_id
            reach_metadata[k]["gage_url"] = f"https://waterdata.usgs.gov/nwis/uv?site_no={gage_id}&legacy=1"

    return reach_metadata



"""General utils used by ripple1d."""

from __future__ import annotations

import glob
import os
from pathlib import Path

import boto3
import geopandas as gpd
import pandas as pd
from dotenv import find_dotenv, load_dotenv
from shapely import Polygon, concave_hull, line_merge, make_valid, union_all
from shapely.geometry import MultiPolygon, Point, Polygon

from ripple1d.errors import (
    RASComputeError,
    RASComputeMeshError,
    RASGeometryError,
    RASStoreAllMapsError,
)
from ripple1d.utils.s3_utils import list_keys

load_dotenv(find_dotenv())


def prj_is_ras(path: str):
    """Verify if prj is from hec-ras model."""
    with open(path) as f:
        prj_contents = f.read()
    if "Proj Title" in prj_contents.split("\n")[0]:
        return True
    else:
        return False


def decode(df: pd.DataFrame):
    """Decode all string columns in a pandas DataFrame."""
    for c in df.columns:
        df[c] = df[c].str.decode("utf-8")
    return df


def get_path(expected_path: str, client: boto3.client = None, bucket: str = None) -> str:
    """Get the path for a file."""
    if client and bucket:
        path = Path(expected_path)
        prefix = path.parent.as_posix().replace("s3:/", "s3://")
        paths = list_keys(client, bucket, prefix, path.suffix)
        if not paths:
            paths = list_keys(client, bucket, prefix, path.suffix.upper())
    else:
        prefix = os.path.dirname(expected_path)
        paths = glob.glob(rf"{prefix}\*{os.path.splitext(expected_path)[1]}")
        if not paths:
            paths = glob.glob(rf"{prefix}\*{os.path.splitext(expected_path)[1].upper()}")

    if expected_path in paths:
        return expected_path
    else:
        for path in paths:
            if path.endswith(Path(expected_path).suffix.upper()):
                return path


def xs_concave_hull(xs: gpd.GeoDataFrame, junction: gpd.GeoDataFrame = None) -> gpd.GeoDataFrame:
    """Compute and return the concave hull (polygon) for a set of cross sections (lines all facing the same direction)."""
    polygons = []
    for river_reach in xs["river_reach"].unique():
        xs_subset = xs[xs["river_reach"] == river_reach]
        points = xs_subset.boundary.explode(index_parts=True).unstack()
        points_last_xs = [Point(coord) for coord in xs_subset["geometry"].iloc[-1].coords]
        points_first_xs = [Point(coord) for coord in xs_subset["geometry"].iloc[0].coords[::-1]]
        polygon = Polygon(points_first_xs + list(points[0]) + points_last_xs + list(points[1])[::-1])
        if isinstance(polygon, MultiPolygon):
            polygons += list(polygon.geoms)
        else:
            polygons.append(polygon)
    if junction is not None:
        for _, j in junction.iterrows():
            polygons.append(junction_hull(xs, j))

    return gpd.GeoDataFrame(
        {"geometry": [union_all([make_valid(p) for p in polygons])]}, geometry="geometry", crs=xs.crs
    )


def determine_junction_xs(xs: gpd.GeoDataFrame, junction: gpd.GeoSeries) -> gpd.GeoDataFrame:
    """Determine the cross sections that bound a junction."""
    junction_xs = []
    for us_river, us_reach in zip(junction.us_rivers.split(","), junction.us_reaches.split(",")):
        xs_us_river_reach = xs[(xs["river"] == us_river) & (xs["reach"] == us_reach)]
        junction_xs.append(
            xs_us_river_reach[xs_us_river_reach["river_station"] == xs_us_river_reach["river_station"].min()]
        )
    for ds_river, ds_reach in zip(junction.ds_rivers.split(","), junction.ds_reaches.split(",")):
        xs_ds_river_reach = xs[(xs["river"] == ds_river) & (xs["reach"] == ds_reach)]
        xs_ds_river_reach["geometry"] = xs_ds_river_reach.reverse()
        junction_xs.append(
            xs_ds_river_reach[xs_ds_river_reach["river_station"] == xs_ds_river_reach["river_station"].max()]
        )
    return pd.concat(junction_xs)


def determine_xs_order(row: gpd.GeoSeries, junction_xs: gpd.gpd.GeoDataFrame):
    """Detemine what order cross sections bounding a junction should be in to produce a valid polygon."""
    candidate_lines = junction_xs[junction_xs["river_reach_rs"] != row["river_reach_rs"]]
    candidate_lines["distance"] = candidate_lines["start"].distance(row.end)
    return candidate_lines.loc[candidate_lines["distance"] == candidate_lines["distance"].min()].geometry.iloc[0]


def junction_hull(xs: gpd.GeoDataFrame, junction: gpd.GeoSeries) -> gpd.GeoDataFrame:
    """Compute and return the concave hull (polygon) for a juction."""
    junction_xs = determine_junction_xs(xs, junction)

    print(type(junction_xs))
    junction_xs["start"] = junction_xs.apply(lambda row: row.geometry.boundary.geoms[0], axis=1)
    junction_xs["end"] = junction_xs.apply(lambda row: row.geometry.boundary.geoms[1], axis=1)
    junction_xs["ordered_lines"] = junction_xs.apply(lambda row: determine_xs_order(row, junction_xs), axis=1)

    coords = []
    for _, row in junction_xs.iterrows():
        coords += list(row["ordered_lines"].coords)
    return Polygon(coords)


def search_contents(lines: list, search_string: str, token: str = "=", expect_one: bool = True) -> list[str]:
    """Split a line by a token and returns the second half of the line if the search_string is found in the first half."""
    results = []
    for line in lines:
        if f"{search_string}{token}" in line:
            results.append(line.split(token)[1])

    if expect_one and len(results) > 1:
        raise ValueError(f"expected 1 result, got {len(results)}")
    elif expect_one and len(results) == 0:
        raise ValueError("expected 1 result, no results found")
    elif expect_one and len(results) == 1:
        return results[0]
    else:
        return results


def replace_line_in_contents(lines: list, search_string: str, replacement: str, token: str = "="):
    """Split a line by a token and replaces the second half of the line (for the first occurence only!)."""
    updated = 0
    for i, line in enumerate(lines):
        if f"{search_string}{token}" in line:
            updated += 1
            lines[i] = line.split(token)[0] + token + replacement
    if updated == 0:
        raise ValueError(f"search_string: {search_string} not found in lines")
    elif updated > 1:
        raise ValueError(f"expected 1 result, got {updated} occurences of {replacement}")
    else:
        return lines


def text_block_from_start_end_str(
    start_str: str, end_strs: list[str], lines: list, additional_lines: int = 0
) -> list[str]:
    """Search for an exact match to the start_str and return all lines from there to a line that contains the end_str."""
    start_str = handle_spaces(start_str, lines)

    start_index = lines.index(start_str)
    end_index = len(lines)
    for line in lines[start_index + 1 :]:
        if end_index != len(lines):
            break
        for end_str in end_strs:
            if end_str in line:
                end_index = lines.index(line) + additional_lines
                break
    return lines[start_index:end_index]


def text_block_from_start_str_to_empty_line(start_str: str, lines: list) -> list[str]:
    """Search for an exact match to the start_str and return all lines from there to the next empty line."""
    start_str = handle_spaces(start_str, lines)
    results = []
    in_block = False
    for line in lines:
        if line == start_str:
            in_block = True
            results.append(line)
            continue

        if in_block:
            if line == "":
                results.append(line)
                return results
            else:
                results.append(line)
    return results


def text_block_from_start_str_length(start_str: str, number_of_lines: int, lines: list) -> list[str]:
    """Search for an exact match to the start token and return a number of lines equal to number_of_lines."""
    start_str = handle_spaces(start_str, lines)
    results = []
    in_block = False
    for line in lines:
        if line == start_str:
            in_block = True
            continue
        if in_block:
            if len(results) >= number_of_lines:
                return results
            else:
                results.append(line)


def data_pairs_from_text_block(lines: list[str], width: int) -> list[tuple[float]]:
    """Split lines at given width to get paired data string. Split the string in half and convert to tuple of floats."""
    pairs = []
    for line in lines:
        for i in range(0, len(line), width):
            x = line[i : int(i + width / 2)]
            y = line[int(i + width / 2) : int(i + width)]
            pairs.append((float(x), float(y)))

    return pairs


def handle_spaces(line: str, lines: list[str]):
    """Handle spaces in the line."""
    if line in lines:
        return line
    elif handle_spaces_arround_equals(line.rstrip(" "), lines):
        return handle_spaces_arround_equals(line.rstrip(" "), lines)
    elif handle_spaces_arround_equals(line + " ", lines) in lines:
        return handle_spaces_arround_equals(line + " ", lines)
    else:
        raise ValueError(f"line: {line} not found in lines")


def handle_spaces_arround_equals(line: str, lines: list[str]) -> str:
    """Handle spaces in the line."""
    if line in lines:
        return line
    elif "= " in line:
        if line.replace("= ", "=") in lines:
            return line.replace("= ", "=")
    else:
        return line.replace("=", "= ")


def assert_no_mesh_error(compute_message_file: str, require_exists: bool):
    """Scan *.computeMsgs.txt for errors encountered. Raise RASComputeMeshError if found."""
    try:
        with open(compute_message_file) as f:
            content = f.read()
    except FileNotFoundError:
        if require_exists:
            raise
    else:
        for line in content.splitlines():
            if "error generating mesh" in line.lower():
                raise RASComputeMeshError(
                    f"'error generating mesh' found in {compute_message_file}. Full file content:\n{content}\n^^^ERROR^^^"
                )


def assert_no_ras_geometry_error(compute_message_file: str):
    """Scan *.computeMsgs.txt for errors encountered."""
    with open(compute_message_file) as f:
        content = f.read()
    for line in content.splitlines():
        if "geometry writer failed" in line.lower() or "error processing geometry" in line.lower():
            raise RASGeometryError(
                f"geometry error found in {compute_message_file}. Full file content:\n{content}\n^^^ERROR^^^"
            )


def assert_no_ras_compute_error_message(compute_message_file: str):
    """Scan *.computeMsgs.txt for errors encountered."""
    with open(compute_message_file) as f:
        content = f.read()
    for line in content.splitlines():
        if "ERROR:" in line:
            raise RASComputeError(
                f"'ERROR:' found in {compute_message_file}. Full file content:\n{content}\n^^^ERROR^^^"
            )


def assert_no_store_all_maps_error_message(compute_message_file: str):
    """Scan *.computeMsgs.txt for errors encountered."""
    with open(compute_message_file) as f:
        content = f.read()
    for line in content.splitlines():
        if "error executing: storeallmaps" in line.lower():
            raise RASStoreAllMapsError(
                f"{repr(line)} found in {compute_message_file}. Full file content:\n{content}\n^^^ERROR^^^"
            )



import json
import logging
import os
import sqlite3
import traceback

import geopandas as gpd
import numpy as np
import pandas as pd
import pyproj
from shapely import Geometry, LineString, MultiLineString, MultiPoint, Point
from shapely.ops import linemerge

from ripple1d.consts import HYDROFABRIC_CRS, METERS_PER_FOOT
from ripple1d.data_model import XS
from ripple1d.ops.subset_gpkg import RippleGeopackageSubsetter
from ripple1d.utils.ripple_utils import xs_concave_hull


class ConflationMetrics:
    """Calculate metrics for a cross section."""

    def __init__(
        self,
        xs_gdf: gpd.GeoDataFrame,
        river_gdf: gpd.GeoDataFrame,
        network_reach: LineString,
        network_reach_plus_ds_reach: LineString,
    ):
        self.xs_gdf = xs_gdf
        self.river_gdf = river_gdf
        self.network_reach = network_reach
        self.network_reach_plus_ds_reach = network_reach_plus_ds_reach
        self.crs = xs_gdf.crs

    def populate_station_elevation(self, row: pd.Series) -> dict:
        """Populate the station elevation for a cross section."""
        xs = XS(row.ras_data.splitlines(), row.river_reach, row.river, row.reach, self.crs)
        return dict(xs.station_elevation_points)

    def populate_thalweg_station(self, row: pd.Series) -> str:
        """Populate the thalweg station for a cross section."""
        df = pd.DataFrame(row["station_elevation"], index=["elevation"]).T
        return df.loc[df["elevation"] == row.thalweg].index[0]

    def thalweg_metrics(self, xs_gdf: gpd.GeoDataFrame) -> dict:
        """Calculate the distance between the thalweg point and the network intersection point."""
        try:
            xs_gdf["station_elevation"] = xs_gdf.apply(lambda row: self.populate_station_elevation(row), axis=1)
            xs_gdf["thalweg_station"] = xs_gdf.apply(lambda row: self.populate_thalweg_station(row), axis=1)
            xs_gdf["thalweg_point"] = xs_gdf.apply(
                lambda row: row.geometry.interpolate(row["thalweg_station"] * METERS_PER_FOOT), axis=1
            )

            xs_gdf["ras_intersection_point"] = None
            for _, r in self.river_gdf.iterrows():
                xs_gdf.loc[xs_gdf["river_reach"] == r["river_reach"], "ras_intersection_point"] = xs_gdf.loc[
                    xs_gdf["river_reach"] == r["river_reach"], :
                ].apply(lambda row: r.geometry.intersection(row.geometry), axis=1)

            xs_gdf["network_intersection_point"] = xs_gdf.apply(
                lambda row: self.network_reach_plus_ds_reach.intersection(row.geometry), axis=1
            )

            xs_gdf["centerline_offset"] = xs_gdf.apply(
                lambda row: row["ras_intersection_point"].distance(row["network_intersection_point"]), axis=1
            )
            xs_gdf["thalweg_offset"] = xs_gdf.apply(
                lambda row: row["thalweg_point"].distance(row["network_intersection_point"]), axis=1
            )

            return {
                "centerline_offset": xs_gdf["centerline_offset"].describe().round().astype(int).to_dict(),
                "thalweg_offset": xs_gdf["thalweg_offset"].describe().round().astype(int).to_dict(),
            }
        except Exception as e:
            logging.error(f"Error: {e}")
            logging.error(f"traceback: {traceback.format_exc()}")

    def ensure_point(self, geom: Geometry):
        """Ensure that the geometry is Point."""
        if isinstance(geom, MultiPoint):
            return list(geom.geoms)[0]
        elif isinstance(geom, Point):
            return geom
        elif isinstance(geom, LineString):
            return list(geom.boundary.geoms)[0]
        else:
            raise ValueError(f"Not a valid geometry type: {type(geom)}. Only Point, MultiPoint,LineString supported.")

    def length_metrics(self, xs_gdf: gpd.GeoDataFrame) -> dict:
        """Calculate the reach length between cross sections along the ras river line and the network reach."""
        try:
            xs_gdf = xs_gdf[
                (xs_gdf["river_station"] == xs_gdf["river_station"].max())
                | (xs_gdf["river_station"] == xs_gdf["river_station"].min())
            ]
            xs_gdf["network_intersection_point"] = xs_gdf.apply(
                lambda row: self.network_reach_plus_ds_reach.intersection(row.geometry), axis=1
            )
            xs_gdf["network_intersection_point"] = xs_gdf.apply(
                lambda row: self.ensure_point(row["network_intersection_point"]), axis=1
            )

            xs_gdf["network_station"] = xs_gdf.apply(
                lambda row: self.network_reach_plus_ds_reach.project(row["network_intersection_point"]), axis=1
            )
            network_length = xs_gdf["network_station"].max() - xs_gdf["network_station"].min()

            if len(xs_gdf["river_reach"].unique()) > 1:
                raise ValueError("Cross sections must all be on the same river reach.")
            else:
                river_line = self.river_gdf.loc[
                    self.river_gdf["river_reach"] == xs_gdf["river_reach"].iloc[0], "geometry"
                ].iloc[0]
                xs_gdf["ras_intersection_point"] = xs_gdf.apply(
                    lambda row: river_line.intersection(row.geometry), axis=1
                )
                xs_gdf["ras_station"] = xs_gdf.apply(
                    lambda row: river_line.project(row["ras_intersection_point"]), axis=1
                )
                ras_length = xs_gdf["ras_station"].max() - xs_gdf["ras_station"].min()

                network_ras_ratio = network_length / ras_length
            return {
                "ras": int(ras_length / METERS_PER_FOOT),
                "network": int(network_length / METERS_PER_FOOT),
                "network_to_ras_ratio": round(float(network_ras_ratio), 2),
            }
        except Exception as e:
            logging.error(f"Error: {e}")
            logging.error(f"traceback: {traceback.format_exc()}")

    # def parrallel_reaches(self, network_reaches: gpd.GeoDataFrame) -> dict:
    #     """Calculate the overlap between the network reach and the cross sections."""
    #     if network_reaches.empty:
    #         return []
    #     overlaps = network_reaches[network_reaches.intersects(self.xs_gdf.union_all())]
    #     if overlaps.empty:
    #         return []
    #     overlaps["overlap"] = overlaps.apply(
    #         lambda row: int(
    #             row["geometry"].intersection(xs_concave_hull(self.xs_gdf)["geometry"].iloc[0]).length / METERS_PER_FOOT
    #         ),
    #         axis=1,
    #     )

    #     return [{"id": str(row["ID"]), "overlap": row["overlap"]} for _, row in overlaps.iterrows()]

    def overlapped_reaches(self, to_reaches: gpd.GeoDataFrame) -> dict:
        """Calculate the overlap between the network reach and the cross sections."""
        if to_reaches.empty:
            return []
        geom_name = to_reaches.geometry.name
        for i, row in to_reaches.iterrows():
            if row[geom_name].intersects(self.xs_gdf.union_all()):
                overlap = (
                    row[geom_name].intersection(xs_concave_hull(self.xs_gdf)["geometry"].iloc[0]).length
                    / METERS_PER_FOOT
                )
                return [{"id": str(row["ID"]), "overlap": int(overlap)}]
        return []

    def eclipsed_reaches(self, network_reaches: gpd.GeoDataFrame) -> dict:
        """Calculate the overlap between the network reach and the cross sections."""
        if network_reaches.empty:
            return []
        eclipsed_reaches = network_reaches[network_reaches.covered_by(xs_concave_hull(self.xs_gdf)["geometry"].iloc[0])]

        return [str(row["ID"]) for _, row in eclipsed_reaches.iterrows()]

    def compute_coverage_metrics(self, xs_gdf: gpd.GeoDataFrame) -> dict:
        """Calculate the coverage metrics for a set of cross sections."""
        try:
            xs_gdf = xs_gdf[
                (xs_gdf["river_station"] == xs_gdf["river_station"].max())
                | (xs_gdf["river_station"] == xs_gdf["river_station"].min())
            ]
            xs_gdf["intersection_point"] = xs_gdf.apply(
                lambda row: self.network_reach_plus_ds_reach.intersection(row.geometry), axis=1
            )
            xs_gdf["intersection_point"] = xs_gdf.apply(
                lambda row: self.ensure_point(row["intersection_point"]), axis=1
            )

            xs_gdf["station_percent"] = xs_gdf.apply(
                lambda row: self.network_reach_plus_ds_reach.project(row["intersection_point"])
                / self.network_reach.length,
                axis=1,
            )
            return {
                "start": float(round(xs_gdf["station_percent"].min(), 2)),
                "end": min([round(xs_gdf["station_percent"].max(), 2), 1]),
            }
        except Exception as e:
            logging.error(f"Error: {e}")
            logging.error(f"traceback: {traceback.format_exc()}")


def compute_conflation_metrics(source_model_directory: str, source_network: str):
    """Compute metrics for a network reach."""
    network_pq_path = source_network["file_name"]
    model_name = os.path.basename(source_model_directory)
    src_gpkg_path = os.path.join(source_model_directory, f"{model_name}.gpkg")
    conflation_json = os.path.join(source_model_directory, f"{model_name}.conflation.json")
    conflation_parameters = json.load(open(conflation_json))

    for network_id in conflation_parameters["reaches"].keys():
        try:
            if conflation_parameters["reaches"][network_id]["eclipsed"] == True:
                continue

            rgs = RippleGeopackageSubsetter(src_gpkg_path, conflation_json, "", network_id)
            layers = {}
            for layer, gdf in rgs.subset_gdfs.items():
                layers[layer] = gdf.to_crs(HYDROFABRIC_CRS)

            network_reaches = gpd.read_parquet(network_pq_path, bbox=layers["XS"].total_bounds)
            network_reach = linemerge(network_reaches.loc[network_reaches["ID"] == int(network_id)].geometry.iloc[0])
            network_reach_plus_ds_reach = combine_reaches(network_reaches, network_id)

            cm = ConflationMetrics(layers["XS"], layers["River"], network_reach, network_reach_plus_ds_reach)

            metrics = {
                "xs": cm.thalweg_metrics(layers["XS"]),
                "lengths": cm.length_metrics(layers["XS"]),
                "coverage": cm.compute_coverage_metrics(layers["XS"]),
            }

            to_id = conflation_parameters["reaches"][network_id]["network_to_id"]
            if to_id in conflation_parameters["reaches"].keys():
                next_to_id = int(conflation_parameters["reaches"][to_id]["network_to_id"])
            else:
                next_to_id = None

            overlapped_reaches = cm.overlapped_reaches(
                network_reaches[network_reaches["ID"].isin([int(to_id), next_to_id])]
            )
            eclipsed_reaches = cm.eclipsed_reaches(network_reaches[network_reaches["ID"] != int(network_id)])

            conflation_parameters["reaches"][network_id].update({"metrics": metrics})
            conflation_parameters["reaches"][network_id].update({"overlapped_reaches": overlapped_reaches})
            conflation_parameters["reaches"][network_id].update({"eclipsed_reaches": eclipsed_reaches})
            conflation_parameters["metadata"]["length_units"] = "feet"
            conflation_parameters["metadata"]["flow_units"] = "cfs"
        except Exception as e:
            logging.error(f"Error: {e}")
            logging.error(f"traceback: {traceback.format_exc()}")
            conflation_parameters["reaches"][network_id].update({"metrics": {}})
    with open(conflation_json, "w") as f:
        f.write(json.dumps(conflation_parameters, indent=4))
    return conflation_parameters


def combine_reaches(network_reaches: gpd.GeoDataFrame, network_id: str) -> LineString:
    """Combine network reaches."""
    reach = network_reaches.loc[network_reaches["ID"] == int(network_id), :]
    to_reach = network_reaches.loc[network_reaches["ID"] == int(reach["to_id"].iloc[0]), :]

    if to_reach.empty:
        return linemerge(reach.geometry.iloc[0])
    else:
        next_to_reach = network_reaches.loc[network_reaches["ID"] == int(to_reach["to_id"].iloc[0]), :]
        if next_to_reach.empty:
            return linemerge(MultiLineString([linemerge(reach.geometry.iloc[0]), linemerge(to_reach.geometry.iloc[0])]))
        else:
            return linemerge(
                MultiLineString(
                    [
                        linemerge(reach.geometry.iloc[0]),
                        linemerge(to_reach.geometry.iloc[0]),
                        linemerge(next_to_reach.geometry.iloc[0]),
                    ]
                )
            )



"""Flask API."""

import time
import traceback
import typing
from http import HTTPStatus

from flask import Flask, Response, jsonify, request
from werkzeug.exceptions import BadRequest

from ripple1d.api import tasks
from ripple1d.api.utils import get_unexpected_and_missing_args
from ripple1d.ops.fim_lib import create_fim_lib, fim_lib_stac, nwm_reach_model_stac
from ripple1d.ops.metrics import compute_conflation_metrics
from ripple1d.ops.ras_conflate import conflate_model
from ripple1d.ops.ras_run import (
    create_model_run_normal_depth,
    run_incremental_normal_depth,
    run_known_wse,
)
from ripple1d.ops.ras_terrain import create_ras_terrain
from ripple1d.ops.subset_gpkg import extract_submodel
from ripple1d.ras_to_gpkg import gpkg_from_ras

app = Flask(__name__)


@app.route("/processes/conflate_model/execution", methods=["POST"])
def process__conflate_model():
    """Enqueue a task to conflate a source model."""
    return enqueue_async_task(conflate_model)


@app.route("/processes/compute_conflation_metrics/execution", methods=["POST"])
def process__compute_conflation_metrics():
    """Enqueue a task to compute conflation metrics."""
    return enqueue_async_task(compute_conflation_metrics)


@app.route("/processes/gpkg_from_ras/execution", methods=["POST"])
def process__gpkg_from_ras():
    """Enqueue a task to create a new GeoPackage from a source model."""
    return enqueue_async_task(gpkg_from_ras)


@app.route("/processes/extract_submodel/execution", methods=["POST"])
def process__extract_submodel():
    """Enqueue a task to create a new GeoPackage for a NWM reach."""
    return enqueue_async_task(extract_submodel)


@app.route("/processes/create_ras_terrain/execution", methods=["POST"])
def process__create_ras_terrain():
    """Enqueue a task to create a new RAS terrain."""
    return enqueue_async_task(create_ras_terrain)


@app.route("/processes/create_model_run_normal_depth/execution", methods=["POST"])
def process__create_model_run_normal_depth():
    """Enqueue a task to calculate the initial normal depth."""
    return enqueue_async_task(create_model_run_normal_depth)


@app.route("/processes/run_incremental_normal_depth/execution", methods=["POST"])
def process__run_incremental_normal_depth():
    """Enqueue a task to calculate the incremental normal depth."""
    return enqueue_async_task(run_incremental_normal_depth)


@app.route("/processes/run_known_wse/execution", methods=["POST"])
def process__run_known_wse():
    """Enqueue a task to calculate the water surface elevation (WSE) based on known inputs."""
    return enqueue_async_task(run_known_wse)


@app.route("/processes/create_fim_lib/execution", methods=["POST"])
def process__create_fim_lib():
    """Enqueue a task to create a FIM library."""
    return enqueue_async_task(create_fim_lib)


@app.route("/processes/nwm_reach_model_stac/execution", methods=["POST"])
def process__nwm_reach_model_stac():
    """Enqueue a task to create a stac item from a fim model."""
    return enqueue_async_task(nwm_reach_model_stac)


@app.route("/processes/fim_lib_stac/execution", methods=["POST"])
def process__fim_lib_stac():
    """Enqueue a task to create a stac item from a fim library."""
    return enqueue_async_task(fim_lib_stac)


@app.route("/ping", methods=["GET"])
def ping():
    """Check the health of the service."""
    return jsonify({"status": "healthy"}), HTTPStatus.OK


@app.route("/processes/test/execution", methods=["POST"])
def test():
    """Test the execution and monitoring of an asynchronous task."""
    response, http_status = enqueue_async_task(tasks.noop)
    if http_status != HTTPStatus.CREATED:
        return jsonify(response.json, HTTPStatus.INTERNAL_SERVER_ERROR)

    timeout_seconds = 10

    start = time.time()
    while time.time() - start < timeout_seconds:
        time.sleep(0.2)
        status = tasks.ogc_status(response.json["jobID"])
        if status == "failed":
            return jsonify({"status": "not healthy"}), HTTPStatus.INTERNAL_SERVER_ERROR
        if status == "successful":
            return jsonify({"status": "healthy"}), HTTPStatus.OK
    return (
        jsonify({"status": f"huey is busy or not active, ping timed out after {timeout_seconds} seconds"}),
        HTTPStatus.GATEWAY_TIMEOUT,
    )


@app.route("/processes/sleep/execution", methods=["POST"])
def process__sleep():
    """Enqueue a task that sleeps for 15 seconds."""
    return enqueue_async_task(tasks.sleep15)


@app.route("/jobs/<task_id>", methods=["GET"])
def get_one_job(task_id):
    """Retrieve OGC status and result for one job.

    Query parameters:
        tb: Choices are ['true', 'false']. Defaults to 'false'. If 'true', the job result's traceback will be included
            in the response, as key 'tb'.
    """
    include_traceback, problem = parse_request_param__bool(param_name="tb", default=False)
    if problem is not None:
        return problem

    task2metadata = tasks.task_status(only_task_id=task_id)

    if len(task2metadata) == 0:
        return jsonify({"type": "process", "detail": f"job ID not found: {task_id}"}), HTTPStatus.NOT_FOUND

    if len(task2metadata) > 1:
        return (
            jsonify({"type": "process", "detail": f"multiple ({len(task2metadata)}) records matched job ID {task_id}"}),
            HTTPStatus.INTERNAL_SERVER_ERROR,
        )

    huey_metadata = task2metadata[task_id]
    return get_ogc_job_metadata_from_huey_metadata(task_id, huey_metadata, include_traceback), HTTPStatus.OK


@app.route("/jobs", methods=["GET"])
def get_all_jobs():
    """Retrieve OGC status and result for all jobs.

    Query parameters:
        tb: Choices are ['true', 'false']. Defaults to 'false'. If 'true', each job result's traceback will be included
            in the response, as key 'tb'.
    """
    include_traceback, problem = parse_request_param__bool(param_name="tb", default=False)
    if problem is not None:
        return problem

    task2metadata = tasks.task_status(only_task_id=None)
    jobs = [
        get_ogc_job_metadata_from_huey_metadata(task_id, huey_metadata, include_traceback)
        for task_id, huey_metadata in task2metadata.items()
    ]
    links = []
    ret = {"jobs": jobs, "links": links}
    return jsonify(ret), HTTPStatus.OK


def parse_request_param__bool(param_name: str, default: bool) -> tuple[bool, tuple]:
    """Get the parameter, assert it is true or false, and return the appropriate Python boolean value as well as a status tuple.

    If there is a problem, the status tuple has two elements: a response message, and a HTTP status to be returned by the endpoint.
    If there is not a problem, the status tuple is actually None.
    """
    arg_tb = request.args.get(param_name)
    if not arg_tb:
        return (default, None)
    elif arg_tb.lower() == "false":
        return (False, None)
    elif arg_tb.lower() == "true":
        return (True, None)
    else:
        return (
            None,
            (
                jsonify(
                    {
                        "type": "process",
                        "detail": f"query param 'tb' should be 'true' or 'false', but got: {repr(arg_tb)}",
                    }
                ),
                HTTPStatus.BAD_REQUEST,
            ),
        )


def get_ogc_job_metadata_from_huey_metadata(task_id: str, huey_metadata: dict, include_traceback: bool) -> dict:
    """Convert huey-style task status metadata into a OGC-style result dictionary."""
    huey_result = tasks.huey.result(task_id, preserve=True)
    if include_traceback is False and huey_result is not None:
        del huey_result["tb"]  # remove the traceback
    ogc_job_metadata = {
        "jobID": task_id,
        "updated": huey_metadata["status_time"],
        "status": huey_metadata["ogc_status"],
        "processID": huey_metadata["func_name"],
        "type": "process",
        "submitter": "",
        "result": huey_result,
    }
    return ogc_job_metadata


@app.route("/jobs/<task_id>", methods=["DELETE"])
def dismiss(task_id):
    """Dismiss a specific task by its ID."""
    try:
        ogc_status = tasks.ogc_status(task_id)
        if ogc_status == "notfound":
            return jsonify({"type": "process", "detail": f"job ID not found: {task_id}"}), HTTPStatus.NOT_FOUND
        elif ogc_status == "accepted":
            tasks.revoke_task(task_id)
            return jsonify({"type": "process", "detail": f"job ID dismissed: {task_id}"}), HTTPStatus.OK
        elif ogc_status == "dismissed":
            return jsonify({"type": "process", "detail": f"job ID dismissed: {task_id}"}), HTTPStatus.OK
        else:
            return (
                jsonify(
                    {
                        "type": "process",
                        "detail": f"failed to dismiss job ID {task_id} due to existing job status '{ogc_status}'",
                    }
                ),
                HTTPStatus.CONFLICT,
            )
    except Exception as e:
        return (
            jsonify({"type": "process", "detail": f"failed to dismiss job ID {task_id} due to internal server error"}),
            HTTPStatus.INTERNAL_SERVER_ERROR,
        )


def enqueue_async_task(func: typing.Callable) -> tuple[Response, HTTPStatus]:
    """Start the execution of the provided func using kwargs from the request body. Assume body is a JSON dictionary."""
    try:
        kwargs = request.json  # can throw BadRequest when parsing body into json
        if not isinstance(kwargs, dict):
            raise BadRequest(f"expected body to be a JSON dictionary, but got: {type(kwargs)}")
    except BadRequest as e:
        return (
            jsonify({"type": "process", "detail": f"could not parse body to json dict. error: {e}"}),
            HTTPStatus.BAD_REQUEST,
        )

    unexpected, missing = get_unexpected_and_missing_args(func, set(kwargs))
    if unexpected or missing:
        return (
            jsonify({"type": "process", "detail": f"unexpected args: {unexpected}, missing args: {missing}"}),
            HTTPStatus.BAD_REQUEST,
        )

    try:
        result = tasks.create_and_enqueue_task(func, kwargs)
    except:
        msg = f"unable to submit task: {traceback.format_exc()}"
        app.logger.error(msg)
        return jsonify({"type": "process", "detail": msg}), HTTPStatus.INTERNAL_SERVER_ERROR
    else:
        return (
            jsonify({"type": "process", "jobID": result.id, "status": "accepted"}),
            HTTPStatus.CREATED,
        )



class ProjectionNotFoundError(Exception):  # noqa: D100
    """TODO."""


class NoDefaultEPSGError(Exception):
    """TODO."""


class ModelNotFoundError(Exception):
    """TODO."""


class NotGeoreferencedError(Exception):
    """TODO."""


class CouldNotIdentifyPrimaryPlanError(Exception):
    """TODO."""


class NoFlowFileSpecifiedError(Exception):
    """TODO."""


class NoGeometryFileSpecifiedError(Exception):
    """TODO."""


class NotAPrjFile(Exception):
    """TODO."""


class NoCRSInferredError(Exception):
    """TODO."""


class UnkownCRSUnitsError(Exception):
    """TODO."""


class HECRASVersionNotInstalledError(Exception):
    """TODO."""


class NoRiverLayerError(Exception):
    """TODO."""


class NoCrossSectionLayerError(Exception):
    """TODO."""


class FlowTitleAlreadyExistsError(Exception):
    """TODO."""


class PlanTitleAlreadyExistsError(Exception):
    """TODO."""


class CouldNotFindAnyPlansError(Exception):
    """TODO."""


class ToManyPlansError(Exception):
    """TODO."""


class RASComputeTimeoutError(Exception):
    """Raised on timeout of API call to Compute_CurrentPlan."""


class RASComputeError(Exception):
    """Raised when *.pNN.computeMsgs.txt indicates error."""


class RASComputeMeshError(Exception):
    """Raised when *.pNN.computeMsgs.txt indicates mesh-specific error."""


class RASGeometryError(Exception):
    """Raised when *.pNN.computeMsgs.txt indicates geometry-specific error."""


class RASStoreAllMapsError(Exception):
    """Raised when *.pNN.computeMsgs.txt indicates StoreAllMaps error (related to RAS Mapper postprocessing)."""


class DepthGridNotFoundError(Exception):
    """Raised when a depth grid is not found when clipping raw RAS output."""


class UnknownVerticalUnits(Exception):
    """Raised when unknown vertical units are specified."""



"""FIM Collection."""

import logging
import os
from datetime import datetime
from pathlib import Path

import pystac
import pystac_client
import pystac_client.errors
from ripple1d.utils.s3_utils import check_s3_key_exists
from ripple1d.utils.stac_utils import (
    key_to_uri,
    upsert_collection,
    upsert_item,
    uri_to_key,
)


class FIMCollection:
    """Class for interacting with a FIM collection in a STAC API."""

    def __init__(self, stac_api: str, collection_id: str, connect: bool = True) -> str:
        self.stac_api = stac_api
        self._collection_id = collection_id
        if connect:
            self.collection = self.load()
        else:
            self.collection = None

    def __repr__(self) -> str:
        """Return a string representation of the FIMCollection."""
        return f"FIMCollection: {self._collection_id}"

    def load(self) -> pystac.Collection:
        """Load the collection from the STAC API."""
        try:
            client = pystac_client.Client.open(self.stac_api)
        except Exception as e:
            raise ConnectionError(f"Error loading STAC API: {self.stac_api}: {e}")

        try:
            return client.get_collection(self._collection_id)
        except Exception:
            raise KeyError(f"Collection `{self._collection_id}` does not exist. Use new_collection() to create.")

    def add_branch_conflation_asset(
        self,
        asset_id: str,
        asset_key: str,
        roles: str = pystac.MediaType.GEOPACKAGE,
        bucket: str = "fim",
    ):
        """Add a branch conflation asset to the collection."""
        asset_url = key_to_uri(asset_key, bucket)
        asset = pystac.Asset(
            href=asset_url,
            description="FIM Branches and Control Nodes",
            roles=[roles],
            extra_fields={"s3_key": asset_key, "updated": datetime.now().isoformat()},
        )
        self.collection.add_asset(asset_id, asset)

    def post_collection_updates(self):
        """Post collection updates to the STAC API."""
        return upsert_collection(self.stac_api, self.collection)

    # def new_item(self, item: pystac.Item):
    #     self.collection.add_item(item)
    #     self.item = item


class FIMCollectionRasItem(FIMCollection):
    """Class for interacting with a FIM collection RAS item in a STAC API."""

    def __init__(
        self,
        stac_api: str,
        collection_id: str,
        item_id: str,
        connect: bool = False,
        load: bool = True,
        local_item: pystac.Item = None,
    ):
        super().__init__(stac_api, collection_id, connect)

        if load:
            if isinstance(local_item, pystac.Item):
                self.item = local_item
                self._item_id = self.item.id
            else:
                self._item_id = item_id
                self.item = self.load_item()
        else:
            raise KeyError(f"Item `{self._item_id}` does not exist. Use new_item() to create.")

    def __repr__(self) -> str:
        """Return a string representation of the FIMCollectionRasItem."""
        return f"FIMCollectionRasItem: {self._collection_id}-{self._item_id}"

    def load_item(self) -> pystac.Item:
        """Load the item from the STAC API."""
        try:
            return self.collection.get_item(self._item_id)
        except Exception:
            raise KeyError(f"Item `{self._item_id}` does not exist. Use new_item() to create.")

    # def add_topo_assets(self, topo_filename: str = "MapTerrain"):

    def post_item_updates(self):
        """Post item updates to the STAC API."""
        return upsert_item(self.stac_api, self.collection.id, self.item)

    def sanitize_ras_stac_props(self):
        """
        TODO: Placeholder function for sanitizing RAS STAC properties.

        Need to update ras-stac and remove this method here or convert to a check.
        """
        # Grab the gpkg location (href)
        for asset_name in self.item.get_assets(role=pystac.MediaType.GEOPACKAGE):
            asset = self.item.assets[asset_name]
            try:
                asset.roles.remove("data")
            except Exception:
                logging.warning(f"no data role: {asset.href}")

            if "ras-geometry-gpkg" not in asset.roles:
                asset.roles.append("ras-geometry-gpkg")

        for asset_name in self.item.get_assets(role="forcing-file"):
            asset = self.item.assets[asset_name]
            try:
                asset.roles.append("hec-ras")
            except Exception:
                logging.warning(f"no data role: {asset.href}")

        for asset_name in self.item.get_assets(role="geometry-file"):
            asset = self.item.assets[asset_name]
            try:
                asset.roles.append("hec-ras")
            except Exception:
                logging.warning(f"no data role: {asset.href}")

        for asset_name in self.item.get_assets(role="plan-file"):
            asset = self.item.assets[asset_name]
            try:
                asset.roles.append("hec-ras")
            except Exception:
                logging.warning(f"no data role: {asset.href}")

        for asset_name in self.item.get_assets(role="projection"):
            asset = self.item.assets[asset_name]
            try:
                asset.roles.remove("projection")
                asset.roles.append("hec-ras")
                asset.roles.append("project-file")
            except Exception:
                logging.warning(f"no data role: {asset.href}")

    def add_s3_key_to_assets(self, bucket: str = "fim"):
        """Add the s3_key to the assets in the item."""
        for asset_name in self.item.assets:
            asset = self.item.assets[asset_name]
            asset.extra_fields["s3_key"] = uri_to_key(asset.href, bucket)

    def ensure_asset_roles_unique(self, bucket: str = "fim"):
        """Ensure that the asset roles are unique."""
        for asset_name in self.item.assets:
            asset = self.item.assets[asset_name]
            asset.roles = list(set(asset.roles))

    def map_topo_assets(
        self,
        date_aquired: str = datetime.now().isoformat(),
        software: str = "ripple1d v0.1.0-alpha.1",
        topo_filename: str = "MapTerrain",
        asset_role: str = "ras-geometry-gpkg",
        source: str = "USGS_Seamless_DEM_13",
    ):
        """
        TODO: Placeholder function for mapping topo assets to FIM collection items.

        This assumes the topo assets are in the same directory as the `ras-geometry-gpkg` asset
        """
        if source != "USGS_Seamless_DEM_13":
            raise NotImplementedError("Only USGS_Seamless_DEM_13 is supported at this time.")

        if asset_role != "ras-geometry-gpkg":
            raise NotImplementedError("Only ras-geometry-gpkg is supported at this time.")

        if topo_filename != "MapTerrain":
            raise NotImplementedError("Only MapTerrain is supported at this time.")

        for asset_name in self.item.get_assets(role=asset_role):
            # logging.debug(asset_name)
            asset_href = self.item.assets[asset_name].href

        topo_href = str(Path(asset_href).parent / "MapTerrain/MapTerrain.ned13.tif").replace("https:/", "https://")
        ras_topo_href = str(Path(asset_href).parent / "MapTerrain/MapTerrain.hdf").replace("https:/", "https://")
        vrt_href = str(Path(asset_href).parent / "MapTerrain/MapTerrain.vrt").replace("https:/", "https://")
        topo_source = "https://rockyweb.usgs.gov/vdelivery/Datasets/Staged/Elevation/13/TIFF/USGS_Seamless_DEM_13.vrt"

        self.item.add_asset(
            f"{topo_filename}.tif",
            pystac.Asset(
                topo_href,
                title=f"{topo_filename}.tif",
                roles=[pystac.MediaType.COG, "ras-topo"],
                extra_fields={
                    "topo_source_url": topo_source,
                    "date_aquired": date_aquired,
                    "software": software,
                },
            ),
        )

        # Add ras topo asset to item
        self.item.add_asset(
            f"{topo_filename}.hdf",
            pystac.Asset(
                ras_topo_href,
                title=f"{topo_filename}.hdf",
                roles=[pystac.MediaType.HDF, "ras-topo"],
                extra_fields={
                    "software": software,
                    "date_aquired": date_aquired,
                },
            ),
        )

        # Add ras topo asset to item
        self.item.add_asset(
            f"{topo_filename}.vrt",
            pystac.Asset(
                vrt_href,
                title=f"{topo_filename}.vrt",
                roles=[pystac.MediaType.XML, "ras-topo"],
                extra_fields={
                    "software": software,
                    "date_aquired": date_aquired,
                },
            ),
        )

    def add_ripple1d_params(
        self,
        date_created: str = datetime.now().isoformat(),
        software: str = "ripple1d v0.1.0-alpha.1",
        asset_role: str = "project-file",
        bucket: str = "fim",
    ) -> bool:
        """
        TODO: Placeholder function for adding ripple1d-params to FIM collection items.

        This assumes the conflation output is in the same directory as the `project-file` asset
        """
        if asset_role != "project-file":
            raise NotImplementedError("Only project-file is supported at this time.")

        for asset_name in self.item.get_assets(role=asset_role):
            asset_href = self.item.assets[asset_name].href

        ripple1d_parameters_href = str((Path(asset_href).parent / "ripple1d_parameters.json")).replace(
            "https:/", "https://"
        )
        ripple1d_parameters_key = uri_to_key(ripple1d_parameters_href, bucket)

        # logging.debug(ripple1d_parameters_key)
        if check_s3_key_exists(bucket, ripple1d_parameters_key):
            self.item.add_asset(
                "ripple1d_parameters.json",
                pystac.Asset(
                    ripple1d_parameters_href,
                    title="ConflationParameters",
                    roles=[pystac.MediaType.JSON, "ripple1d-params"],
                    extra_fields={
                        "software": software,
                        "date_created": date_created,
                    },
                ),
            )
            return True
        else:
            return False


class FIMCollectionRasDGItem(FIMCollection):
    """
    Class for interacting with a FIM collection RAS Depth Grid item in a STAC API.

    Not Implemented
    """

    def __init__(self, stac_api: str, collection_id: str, item_id: str, load: bool = True):
        super().__init__(stac_api, collection_id, load)
        raise NotImplementedError("FIMCollectionRasDGItem is not implemented.")
        self._item_id = item_id
        if load:
            self.item = self.load_item()
        else:
            raise KeyError(f"Item `{self._item_id}` does not exist. Use new_item() to create.")

    def __repr__(self) -> str:
        """Return a string representation of the FIMCollectionRasDGItem."""
        return f"FIMCollectionDGItem: {self._collection_id}-{self._item_id}"

    def load_item(self) -> pystac.Item:
        """Load the item from the STAC API."""
        try:
            return self.collection.get_item(self._item_id)
        except Exception:
            raise KeyError(f"Item `{self._item_id}` does not exist. Use new_item() to create.")
